"""Defines class Project."""

import pickle
import codecs
import os
from typing import List, Optional, Tuple, TypedDict, Union, cast, overload

import reapy_boost
from reapy_boost import reascript_api as RPR
from reapy_boost.core import ReapyObject
from reapy_boost.errors import RedoError, UndoError


class MeasureInfo(TypedDict):
    """
    start: float
        in quarter notes
    end: float
        in quarter notes
    num: int
        numenator of time signature (e.g. 3 in 3/8)
    denom: int
        denominator of time signature (e.g. 8 in 3/8)
    bpm: float
    """
    start: float
    end: float
    num: int
    denom: int
    bpm: float


class Project(ReapyObject):
    """REAPER project."""

    def __init__(self, id: Optional[Union[str, int]] = None, index: int = -1):
        """
        Build project either by ID or index.

        Parameters
        ----------
        id : None, str or int, optional
            Project identifier.
            When None (default), `index is used instead.
            An integer is interpreted as the project index in GUI.
            A string starting with '(ReaProject*)0x' is interpreted
            as a ReaScript identifier.
            Otherwise, `id` is the project name. In that case, the .rpp
            extension is optional.
        index : int, optional
            Project index in GUI (default=-1, corresponds to current
            project).
        """
        if isinstance(id, int):
            id, index = None, cast(int, id)
        if id is None:
            id = cast(str, RPR.EnumProjects(index, None, 0)[0])  # type:ignore
        if not id.startswith('(ReaProject*)0x'):
            id = Project._from_name(id).id
        self.id: str = id
        self._filename = None

    def __eq__(self, other: object) -> bool:
        if hasattr(other, 'id'):
            return self.id == other.id
        return False

    @property
    def _args(self) -> Tuple[str]:
        return self.id,

    @staticmethod
    def _from_name(name: str) -> 'Project':
        """Return project with corresponding name.

        Parameters
        ----------
        name : str
            Project file name. Including the extension ('.rpp')
            is optional.

        Returns
        -------
        Project

        Raises
        ------
        NameError
            If no project with the corresponding name is open.
        """
        if not name.lower().endswith('.rpp'):
            name += '.rpp'
        with reapy_boost.inside_reaper():
            for project in reapy_boost.get_projects():
                project_name = project.name[:-4] + '.rpp'
                if project_name == name:
                    return project
        raise NameError('"{}" is not currently open.'.format(name))

    @reapy_boost.inside_reaper()
    def _get_track_by_name(self, name: str) -> 'reapy_boost.Track':
        """Return first track with matching name."""
        for track in self.tracks:
            if track.name == name:
                return track
        raise KeyError(name)

    def add_marker(
        self,
        position: float,
        name: str = "",
        color: Union[int, Tuple[int, int, int]] = 0
    ) -> 'reapy_boost.Marker':
        """
        Create new marker and return its index.

        Parameters
        ----------
        position : float
            Marker position in seconds.
        name : str, optional
            Marker name.
        color : int or tuple of int, optional
            Marker color. Integers correspond to REAPER native colors.
            Tuple must be RGB triplets of integers between 0 and 255.

        Returns
        -------
        marker : reapy_boost.Marker
            New marker.

        Notes
        -----
        If a marker with the same position and name already exists, no
        new marker will be created, and the existing marker index will
        be returned.
        """
        if isinstance(color, tuple):
            color = reapy_boost.rgb_to_native(color) | 0x1000000
        marker_id = RPR.AddProjectMarker2(  # type:ignore
            self.id, False, position, 0, name, -1, color
        )
        marker = reapy_boost.Marker(self, index=marker_id)
        return marker

    def add_region(
        self,
        start: float,
        end: float,
        name: str = "",
        color: Union[int, Tuple[int, int, int]] = 0
    ) -> 'reapy_boost.Region':
        """
        Create new region and return its index.

        Parameters
        ----------
        start : float
            Region start in seconds.
        end : float
            Region end in seconds.
        name : str, optional
            Region name.
        color : int or tuple of int, optional
            Region color. Integers correspond to REAPER native colors.
            Tuple must be RGB triplets of integers between 0 and 255.

        Returns
        -------
        region : reapy_boost.Region
            New region.
        """
        if isinstance(color, tuple):
            color = reapy_boost.rgb_to_native(color) | 0x1000000
        region_id = RPR.AddProjectMarker2(  # type:ignore
            self.id, True, start, end, name, -1, color
        )
        region = reapy_boost.Region(self, index=region_id)
        return region

    @reapy_boost.inside_reaper()
    def add_track(self, index: int = 0, name: str = "") -> 'reapy_boost.Track':
        """
        Add track at a specified index.

        Parameters
        ----------
        index : int
            Index at which to insert track.
        name : str, optional
            Name of created track.

        Returns
        -------
        track : Track
            New track.
        """
        n_tracks = self.n_tracks
        index = max(-n_tracks, min(index, n_tracks))
        if index < 0:
            index = index % n_tracks
        with self.make_current_project():
            RPR.InsertTrackAtIndex(index, True)  # type:ignore
        track = self.tracks[index]
        track.name = name
        return track

    @property
    def any_track_solo(self) -> bool:
        """
        Test whether any track is soloed in project.

        Returns
        -------
        any_track_solo : bool
            Whether any track is soloed in project.
        """
        any_track_solo = bool(RPR.AnyTrackSolo(self.id))  # type:ignore
        return any_track_solo

    def beats_to_measures(self, beats: float) -> Tuple[int, float, float]:
        """Get measure from quarter notes.

        Parameters
        ----------
        beats : float
            quarter notes from start

        Returns
        -------
        Tuple[int, float, float]
            measure number: int
            measure start in quarter notes: float
            measure end in quarter notes: float
        """
        (m, _, _, m_start, m_end
         ) = RPR.TimeMap_QNToMeasures(  # type:ignore
             self.id, beats, 0, 0
         )
        return m, m_start, m_end

    def beats_to_time(self, beats: float) -> float:
        """
        Convert beats to time in seconds.

        Parameters
        ----------
        beats : float
            Time in beats

        Returns
        -------
        time : float
            Converted time in seconds.

        See also
        --------
        Project.time_to_beats
        """
        time = RPR.TimeMap2_QNToTime(self.id, beats)  # type:ignore
        return time

    def begin_undo_block(self) -> None:
        """
        Start a new undo block.
        """
        RPR.Undo_BeginBlock2(self.id)  # type:ignore

    @property
    def bpi(self) -> int:
        """
        Return project BPI (numerator of time signature).

        Returns
        -------
        bpi : int
            Numerator of time signature.
        """
        return self.time_signature[1]

    @property
    def bpm(self) -> float:
        """
        Project BPM (beats per minute).

        :type: float
        """
        return self.time_signature[0]

    @bpm.setter
    def bpm(self, bpm: float) -> None:
        """
        Set project BPM (beats per minute).

        Parameters
        ----------
        bpm : float
            Tempo in beats per minute.
        """
        RPR.SetCurrentBPM(self.id, bpm, True)  # type:ignore

    @property
    def buffer_position(self) -> float:
        """
        Position of next audio block being processed in seconds.

        :type: float

        See also
        --------
        Project.play_position
            Latency-compensated actual-what-you-hear position.
        """
        return RPR.GetPlayPosition2Ex(self.id)  # type:ignore

    @reapy_boost.inside_reaper()
    def bypass_fx_on_all_tracks(self, bypass: bool = True) -> None:
        """
        Bypass or un-bypass FX on all tracks.

        Parameters
        ----------
        bypass : bool
            Whether to bypass or un-bypass FX.
        """
        with self.make_current_project():
            RPR.BypassFxAllTracks(bypass)  # type:ignore

    @property
    def can_redo(self) -> bool:
        """
        Whether redo is possible.

        :type: bool
        """
        try:
            RPR.Undo_CanRedo2(self.id)  # type:ignore
            can_redo = True
        except AttributeError:  # Bug in ReaScript API
            can_redo = False
        return can_redo

    @property
    def can_undo(self) -> bool:
        """
        Whether undo is possible.

        :type: bool
        """
        try:
            RPR.Undo_CanUndo2(self.id)  # type:ignore
            can_undo = True
        except AttributeError:  # Bug in ReaScript API
            can_undo = False
        return can_undo

    def close(self) -> None:
        """Close project and its correspondig tab."""
        self._filename = os.path.join(self.path, self.name)
        with self.make_current_project():
            reapy_boost.perform_action(40860)

    @property
    def cursor_position(self) -> float:
        """
        Edit cursor position in seconds.

        :type: float
        """
        position = RPR.GetCursorPositionEx(self.id)  # type:ignore
        return position

    @cursor_position.setter
    def cursor_position(self, position: float) -> None:
        """
        Set edit cursor position.

        Parameters
        ----------
        position : float
            New edit cursor position in seconds.
        """
        RPR.SetEditCurPos(position, True, True)  # type:ignore

    @reapy_boost.inside_reaper()
    def disarm_rec_on_all_tracks(self) -> None:
        """
        Disarm record on all tracks.
        """
        with self.make_current_project():
            RPR.ClearAllRecArmed()  # type:ignore

    def end_undo_block(self, description: str = "", flags: int = -1) -> None:
        """
        End undo block.

        Parameters
        ----------
        description : str
            Undo block description.
        flags : int
            1: track configurations
            2: track FX
            4: track items
            8: project states
            16: freeze states
        """
        RPR.Undo_EndBlock2(self.id, description, flags)  # type:ignore

    @property
    def focused_fx(self) -> Optional['reapy_boost.FX']:
        """
        FX that has focus if any, else None.

        :type: FX or NoneType
        """
        return self._focused_fx_inside()

    @reapy_boost.inside_reaper()
    def _focused_fx_inside(self) -> Optional['reapy_boost.FX']:
        if not self.is_current_project:
            return None
        res = RPR.GetFocusedFX(0, 0, 0)  # type:ignore
        if not res[0]:
            return None
        if res[1] == 0:
            track = self.master_track
        else:
            track = self.tracks[res[1] - 1]
        if res[0] == 1:  # Track FX
            return track.fxs[res[3]]
        # Take FX
        item = track.items[res[2]]
        take = item.takes[res[3] // 2**16]
        return take.fxs[res[3] % 2**16]

    def get_info_string(self, param_name: str) -> str:
        """
        Parameters
        ----------
        param_name : str
            MARKER_GUID:X : get the GUID (unique ID) of the marker or region
                with index X, where X is the index passed to
                EnumProjectMarkers, not necessarily the displayed number
            RECORD_PATH :
                recording directory -- may be blank or a relative path,
                to get the effective path see GetProjectPathEx()
            RENDER_FILE : render directory
            RENDER_PATTERN : render file name (may contain wildcards)
            RENDER_FORMAT : base64-encoded sink configuration
                (see project files, etc). Callers can also pass a simple
                4-byte string (non-base64-encoded), e.g. "evaw" or "l3pm",
                to use default settings for that sink type.
            RENDER_FORMAT2 : base64-encoded secondary sink configuration.
                Callers can also pass a simple 4-byte string (non-base64-encoded),
                e.g. "evaw" or "l3pm", to use default settings for
                that sink type, or "" to disable secondary render.
                Formats available on this machine:
                "wave" "aiff" "iso " "ddp " "flac" "mp3l" "oggv" "OggS"
                "FFMP" "GIF " "LCF " "wvpk"
        """
        _, _, _, result, _ = RPR.GetSetProjectInfo_String(  # type:ignore
            self.id, param_name, 'valuestrNeedBig', False
        )
        return result  # type:ignore

    def get_info_value(self, param_name: str) -> float:
        """
        Parameters
        ----------
        param_name : str
            RENDER_SETTINGS : &(1|2)=0:master mix, &1=stems+master mix,
                &2=stems only, &4=multichannel tracks to multichannel files,
                &8=use render matrix, &16=tracks with only mono media
                to mono files, &32=selected media items,
                &64=selected media items via master
            RENDER_BOUNDSFLAG : 0=custom time bounds, 1=entire project,
                2=time selection, 3=all project regions,
                4=selected media items, 5=selected project regions
            RENDER_CHANNELS : number of channels in rendered file
            RENDER_SRATE : sample rate of rendered file
                (or 0 for project sample rate)
            RENDER_STARTPOS : render start time when RENDER_BOUNDSFLAG=0
            RENDER_ENDPOS : render end time when RENDER_BOUNDSFLAG=0
            RENDER_TAILFLAG : apply render tail setting when rendering:
                &1=custom time bounds, &2=entire project, &4=time selection,
                &8=all project regions, &16=selected media items,
                &32=selected project regions
            RENDER_TAILMS : tail length in ms to render
                (only used if RENDER_BOUNDSFLAG and RENDER_TAILFLAG are set)
            RENDER_ADDTOPROJ : 1=add rendered files to project
            RENDER_DITHER : &1=dither, &2=noise shaping, &4=dither stems,
                &8=noise shaping on stems
            PROJECT_SRATE : samplerate (ignored unless PROJECT_SRATE_USE set)
            PROJECT_SRATE_USE : set to 1 if project samplerate is used
        """
        return RPR.GetSetProjectInfo(  # type:ignore
            self.id, param_name, 0, False
        )

    def get_play_rate(self, position: float) -> float:
        """
        Return project play rate at a given position.

        Parameters
        ----------
        position : float
            Position in seconds.

        Returns
        -------
        play_rate : float
            Play rate at the given position.

        See also
        --------
        Project.play_rate
            Project play rate at the current position.
        """
        play_rate = RPR.Master_GetPlayRateAtTime(  # type:ignore
            position, self.id)
        return play_rate  # type:ignore

    def get_selected_item(self, index: int) -> 'reapy_boost.Item':
        """
        Return index-th selected item.

        Parameters
        ----------
        index : int
            Item index.

        Returns
        -------
        item : Item
            index-th selected item.
        """
        item_id = RPR.GetSelectedMediaItem(self.id, index)  # type:ignore
        item = reapy_boost.Item(item_id)
        return item

    def get_selected_track(self, index: int) -> 'reapy_boost.Track':
        """
        Return index-th selected track.

        Parameters
        ----------
        index : int
            Track index.

        Returns
        -------
        track : Track
            index-th selected track.
        """
        track_id = RPR.GetSelectedTrack(self.id, index)  # type:ignore
        track = reapy_boost.Track(track_id)
        return track

    @reapy_boost.inside_reaper()
    def get_track_by_guid(self, guid_string: str) -> 'reapy_boost.Track':
        """
        Get track with giver GUID string {xyz-...}.

        Parameters
        ----------
        guid_string : str

        Returns
        -------
        Track

        Raises
        ------
        KeyError
            If no track with the guid string found in project
        """
        for tr in self.tracks:
            if guid_string == tr.GUID:
                return tr
        raise KeyError(guid_string)

    def get_ext_state(self,
                      section: str,
                      key: str,
                      pickled: bool = False) -> Union[str, object]:
        """
        Return external state of project.

        Parameters
        ----------
        section : str
        key : str
        pickled: bool
            Whether data was pickled or not.

        Returns
        -------
        value : str
            If key or section does not exist an empty string is returned.
        """
        value = RPR.GetProjExtState(  # type:ignore
            self.id,
            section,
            key,
            "",
            2**31 - 1
        )[4]
        if value and pickled:
            value = pickle.loads(codecs.decode(value.encode(), "base64"))
        return value  # type:ignore

    def glue_items(self, within_time_selection: bool = False) -> None:
        """
        Glue items (action shortcut).

        Parameters
        ----------
        within_time_selection : bool
            If True, glue items within time selection.
        """
        action_id = 41588 if within_time_selection else 40362
        self.perform_action(action_id)

    @property
    def has_valid_id(self) -> bool:
        """
        Whether ReaScript ID is still valid.

        For instance, if project has been closed, ID will not be valid
        anymore.

        :type: bool
        """
        return bool(
            RPR.ValidatePtr(  # type:ignore
                *self._get_pointer_and_name()
            )
        )

    @property
    def is_dirty(self) -> bool:
        """
        Whether project is dirty (i.e. needing save).

        :type: bool
        """
        is_dirty = RPR.IsProjectDirty(self.id)  # type:ignore
        return bool(is_dirty)

    @property
    def is_current_project(self) -> bool:
        """
        Whether project is current project.

        :type: bool
        """
        is_current = self == Project()
        return is_current

    @property
    def is_paused(self) -> bool:
        """
        Return whether project is paused.

        :type: bool
        """
        return bool(RPR.GetPlayStateEx(self.id) & 2)  # type:ignore

    @property
    def is_playing(self) -> bool:
        """
        Return whether project is playing.

        :type: bool
        """
        return bool(RPR.GetPlayStateEx(self.id) & 1)  # type:ignore

    @property
    def is_recording(self) -> bool:
        """
        Return whether project is recording.

        :type: bool
        """
        return bool(RPR.GetPlayStateEx(self.id) & 4)  # type:ignore

    @property
    def is_stopped(self) -> bool:
        """
        Return whether project is stopped.

        :type: bool
        """
        return self._is_stopped_inside()

    @reapy_boost.inside_reaper()
    def _is_stopped_inside(self) -> bool:
        return not self.is_playing and not self.is_paused

    @property
    def items(self) -> List['reapy_boost.Item']:
        """
        List of items in project.

        :type: list of Item
        """
        return self._items_inside()

    @reapy_boost.inside_reaper()
    def _items_inside(self) -> List['reapy_boost.Item']:
        n_items = self.n_items
        item_ids = [
            RPR.GetMediaItem(  #type:ignore
                self.id, i
            ) for i in range(n_items)
        ]
        return list(map(reapy_boost.Item, item_ids))

    @property
    def length(self) -> float:
        """
        Project length in seconds.

        :type: float
        """
        length = RPR.GetProjectLength(self.id)  #type:ignore
        return length

    @property
    def last_touched_fx(
        self
    ) -> Tuple[Optional['reapy_boost.FX'], Optional[int]]:
        """
        Last touched FX and corresponding parameter index.

        :type: FX, int or NoneType, NoneType

        Notes
        -----
        Only Track FX are detected by this property. If last touched
        FX is a Take FX, this property is ``(None, None)``.

        Examples
        --------
        >>> fx, index = project.last_touched_fx
        >>> fx.name
        'VSTi: ReaSamplOmatic5000 (Cockos)'
        >>> fx.params[index].name
        "Volume"
        """
        return self._last_touched_fx_inside()

    @reapy_boost.inside_reaper()
    def _last_touched_fx_inside(
        self
    ) -> Tuple[Optional['reapy_boost.FX'], Optional[int]]:
        if not self.is_current_project:
            fx, index = None, None
        else:
            res = RPR.GetLastTouchedFX(0, 0, 0)
            if not res[0]:
                fx, index = None, None
            else:
                if res[1]:
                    track = self.tracks[res[1] - 1]
                else:
                    track = self.master_track
                fx, index = track.fxs[res[2]], res[3]
        return fx, index

    @property
    def loop_points(self) -> Tuple[float, float]:
        _, _, _, startOut, endOut, _ = RPR.GetSet_LoopTimeRange2(  # type:ignore
            self.id, False, False, False, False, False
        )
        return startOut, endOut

    @loop_points.setter
    def loop_points(self, points: Tuple[float, float]) -> None:
        RPR.GetSet_LoopTimeRange2(  # type:ignore
            self.id, True, True, points[0], points[1], False
        )

    def make_current_project(self) -> '_MakeCurrentProject':
        """
        Set project as current project.

        Can also be used as a context manager to temporarily set
        the project as current project and then go back to the original
        situation.

        Examples
        --------
        >>> p1 = reapy_boost.Project()  # current project
        >>> p2 = reapy_boost.Project(1)  # other project
        >>> p2.make_current_project()  # now p2 is current project
        >>> with p1.make_current_project():
        ...     do_something()  # current project is temporarily p1
        >>> # and p2 is current project again
        """
        return _MakeCurrentProject(self)

    def mark_dirty(self) -> None:
        """
        Mark project as dirty (i.e. needing save).
        """
        RPR.MarkProjectDirty(self.id)  # type:ignore

    @property
    def markers(self) -> List['reapy_boost.Marker']:
        """
        List of project markers.

        :type: list of reapy_boost.Marker
        """
        return self._markers_inside()

    @reapy_boost.inside_reaper()
    def _markers_inside(self) -> List['reapy_boost.Marker']:
        ids = [
            RPR.EnumProjectMarkers2(  # type:ignore
                self.id, i, 0, 0, 0, 0, 0
            ) for i in range(self.n_regions + self.n_markers)
        ]
        return [
            reapy_boost.Marker(self, index=i[7], enum_index=i[0] - 1)
            for i in ids if not i[3]
        ]

    @property
    def master_track(self) -> 'reapy_boost.Track':
        """
        Project master track.

        :type: Track
        """
        track_id = RPR.GetMasterTrack(self.id)  # type:ignore
        master_track = reapy_boost.Track(track_id)
        return master_track

    def measure_info(self, measure: int) -> MeasureInfo:
        """Information about measure.

        Parameters
        ----------
        measure : int
            number from the project start

        Returns
        -------
        MeasureInfo:
            start: float
                in quarter notes
            end: float
                in quarter notes
            num: int
                numenator of time signature (e.g. 3 in 3/8)
            denom: int
                denominator of time signature (e.g. 8 in 3/8)
            bpm: float
        """
        (_, _, _, start, end, num, denom,
            bpm) = RPR.TimeMap_GetMeasureInfo(  # type:ignore
            self.id, measure-1, 0, 0, 0, 0, 0)
        return MeasureInfo(
            start=cast(float, start),
            end=cast(float, end),
            num=int(num),
            denom=int(denom),
            bpm=cast(float, bpm),
        )

    @reapy_boost.inside_reaper()
    def mute_all_tracks(self, mute: bool = True) -> None:
        """
        Mute or unmute all tracks.

        Parameters
        ----------
        mute : bool, optional
            Whether to mute or unmute all tracks (default=True).

        See also
        --------
        Project.unmute_all_tracks
        """
        with self.make_current_project():
            RPR.MuteAllTracks(mute)  # type:ignore

    @property
    def n_items(self) -> int:
        """
        Number of items in project.

        :type: int
        """
        n_items = RPR.CountMediaItems(self.id)  #type:ignore
        return cast(int, n_items)

    @property
    def n_markers(self) -> int:
        """
        Number of markers in project.

        :type: int
        """
        n_markers = RPR.CountProjectMarkers(self.id, 0, 0)[2]  #type:ignore
        return cast(int, n_markers)

    @property
    def n_regions(self) -> int:
        """
        Number of regions in project.

        :type: int
        """
        n_regions = RPR.CountProjectMarkers(self.id, 0, 0)[3]  #type:ignore
        return cast(int, n_regions)

    @property
    def n_selected_items(self) -> int:
        """
        Number of selected media items.

        :type: int
        """
        n_items = RPR.CountSelectedMediaItems(self.id)  #type:ignore
        return cast(int, n_items)

    @property
    def n_selected_tracks(self) -> int:
        """
        Number of selected tracks in project (excluding master).

        :type: int
        """
        n_tracks = RPR.CountSelectedTracks2(self.id, False)  #type:ignore
        return cast(int, n_tracks)

    @property
    def n_tempo_markers(self) -> int:
        """
        Number of tempo/time signature markers in project.

        :type: int
        """
        n_tempo_markers = RPR.CountTempoTimeSigMarkers(self.id)  #type:ignore
        return cast(int, n_tempo_markers)

    @property
    def n_tracks(self) -> int:
        """
        Number of tracks in project.

        :type: int
        """
        n_tracks = RPR.CountTracks(self.id)  #type:ignore
        return cast(int, n_tracks)

    @property
    def name(self) -> str:
        """
        Project name.

        :type: str
        """
        _, name, _ = RPR.GetProjectName(self.id, "", 2048)  #type:ignore
        return cast(str, name)

    def open(self, in_new_tab: bool = False) -> None:
        """
        Open project, if it was closed by Project.close.

        Parameters
        ----------
        in_new_tab : bool, optional
            whether should be opened in new tab

        Raises
        ------
        RuntimeError
            If hasn't been closed by Project.close yet
        """
        if self._filename is None:
            raise RuntimeError("project hasn't been closed")
        self.id = reapy_boost.open_project(self._filename, in_new_tab).id

    def pause(self) -> None:
        """
        Hit pause button.
        """
        RPR.OnPauseButtonEx(self.id)  #type:ignore

    @property
    def path(self) -> str:
        """
        Project path.

        :type: str
        """
        _, path, _ = RPR.GetProjectPathEx(self.id, "", 2048)  #type:ignore
        return cast(str, path)

    def perform_action(self, action_id: int) -> None:
        """
        Perform action with ID `action_id` in the main Actions section.

        Parameters
        ----------
        action_id : int
            Action ID in the main Actions section.
        """
        RPR.Main_OnCommandEx(action_id, 0, self.id)  #type:ignore

    def play(self) -> None:
        """
        Hit play button.
        """
        RPR.OnPlayButtonEx(self.id)  #type:ignore

    @property
    def play_position(self) -> float:
        """
        Latency-compensated actual-what-you-hear position in seconds.

        :type: float

        See also
        --------
        Project.buffer_position
            Position of next audio block being processed.
        """
        return RPR.GetPlayPositionEx(self.id)  #type:ignore

    @property
    def play_rate(self) -> float:
        """
        Project play rate at the cursor position.

        :type: float

        See also
        --------
        Project.get_play_rate
            Return project play rate at a specified time.
        """
        play_rate = RPR.Master_GetPlayRate(self.id)  #type:ignore
        return cast(float, play_rate)

    @reapy_boost.inside_reaper()
    def record(self) -> None:
        """Hit record button."""
        with self.make_current_project():
            reapy_boost.perform_action(1013)

    def redo(self) -> None:
        """
        Redo last action.

        Raises
        ------
        RedoError
            If impossible to redo.
        """
        success = RPR.Undo_DoRedo2(self.id)  #type:ignore
        if not success:
            raise RedoError

    @property
    def regions(self) -> List['reapy_boost.Region']:
        """
        List of project regions.

        :type: list of reapy_boost.Region
        """
        return self._regions_inside()

    @reapy_boost.inside_reaper()
    def _regions_inside(self) -> List['reapy_boost.Region']:
        ids = [
            RPR.EnumProjectMarkers2(  # type:ignore
                self.id, i, 0, 0, 0, 0, 0
            ) for i in range(self.n_regions + self.n_markers)
        ]
        return [
            reapy_boost.Region(self, index=i[7], enum_index=i[0] - 1)
            for i in ids if i[3]
        ]

    def save(self, force_save_as: bool = False) -> None:
        """
        Save project.

        Parameters
        ----------
        force_save_as : bool
            Force using "Save as" instead of "Save".
        """
        RPR.Main_SaveProject(self.id, force_save_as)  # type:ignore

    def select(
        self,
        start: float,
        end: Optional[float] = None,
        length: Optional[float] = None
    ) -> None:
        if end is None:
            message = "Either `end` or `length` must be specified."
            assert length is not None, message
            end = start + length
        self.time_selection = start, end

    def select_all_items(self, selected: bool = True) -> None:
        """
        Select or unselect all items, depending on `selected`.

        Parameters
        ----------
        selected : bool
            Whether to select or unselect items.
        """
        RPR.SelectAllMediaItems(self.id, selected)  # type:ignore

    def select_all_tracks(self) -> None:
        """Select all tracks."""
        self.perform_action(40296)

    @property
    def selected_envelope(self) -> Optional['reapy_boost.Envelope']:
        """
        Project selected envelope.

        :type: reapy_boost.Envelope or None
        """
        return self._selected_envelope_inside()

    @reapy_boost.inside_reaper()
    def _selected_envelope_inside(self) -> Optional['reapy_boost.Envelope']:
        selected_track = self.get_selected_track(0)
        envelope_id = cast(
            str,
            RPR.GetSelectedTrackEnvelope(self.id)  # type:ignore
        )
        envelope = None if envelope_id == 0 else reapy_boost.Envelope(
            selected_track, envelope_id
        )
        return envelope

    @property
    def selected_items(self) -> List['reapy_boost.Item']:
        """
        List of all selected items.

        :type: list of Item

        See also
        --------
        Project.get_selected_item
            Return a specific selected item.
        """
        return self._selected_items_inside()

    @reapy_boost.inside_reaper()
    def _selected_items_inside(self) -> List['reapy_boost.Item']:
        return [
            reapy_boost.Item(
                RPR.GetSelectedMediaItem(  # type:ignore
                    self.id, i
                )
            ) for i in range(self.n_selected_items)
        ]

    @property
    def selected_tracks(self) -> List['reapy_boost.Track']:
        """
        List of selected tracks (excluding master).

        :type: list of Track
        """
        return self._selected_tracks_inside()

    @selected_tracks.setter
    def selected_tracks(self, tracks: List['reapy_boost.Track']) -> None:
        self.unselect_all_tracks()
        for track in tracks:
            track.select()

    @reapy_boost.inside_reaper()
    def _selected_tracks_inside(self) -> List['reapy_boost.Track']:
        return [
            reapy_boost.Track(
                RPR.GetSelectedTrack(  # type:ignore
                    self.id, i
                )
            ) for i in range(self.n_selected_tracks)
        ]

    def set_info_string(self, param_name: str, param_string: str) -> None:
        """
        Parameters
        ----------
        param_name : str
            MARKER_GUID:X : get the GUID (unique ID) of the marker or region
                with index X, where X is the index passed to
                EnumProjectMarkers, not necessarily the displayed number
            RECORD_PATH :
                recording directory -- may be blank or a relative path,
                to get the effective path see GetProjectPathEx()
            RENDER_FILE : render directory
            RENDER_PATTERN : render file name (may contain wildcards)
            RENDER_FORMAT : base64-encoded sink configuration
                (see project files, etc). Callers can also pass a simple
                4-byte string (non-base64-encoded), e.g. "evaw" or "l3pm",
                to use default settings for that sink type.
            RENDER_FORMAT2 : base64-encoded secondary sink configuration.
                Callers can also pass a simple 4-byte string (non-base64-encoded),
                e.g. "evaw" or "l3pm", to use default settings for
                that sink type, or "" to disable secondary render.
                Formats available on this machine:
                "wave" "aiff" "iso " "ddp " "flac" "mp3l" "oggv" "OggS"
                "FFMP" "GIF " "LCF " "wvpk"
        param_string : str
        """
        RPR.GetSetProjectInfo_String(  # type:ignore
            self.id, param_name, param_string, True)

    def set_info_value(self, param_name: str, param_value: float) -> None:
        """
        Parameters
        ----------
        param_name : str
            RENDER_SETTINGS : &(1|2)=0:master mix, &1=stems+master mix,
                &2=stems only, &4=multichannel tracks to multichannel files,
                &8=use render matrix, &16=tracks with only mono media
                to mono files, &32=selected media items,
                &64=selected media items via master
            RENDER_BOUNDSFLAG : 0=custom time bounds, 1=entire project,
                2=time selection, 3=all project regions,
                4=selected media items, 5=selected project regions
            RENDER_CHANNELS : number of channels in rendered file
            RENDER_SRATE : sample rate of rendered file
                (or 0 for project sample rate)
            RENDER_STARTPOS : render start time when RENDER_BOUNDSFLAG=0
            RENDER_ENDPOS : render end time when RENDER_BOUNDSFLAG=0
            RENDER_TAILFLAG : apply render tail setting when rendering:
                &1=custom time bounds, &2=entire project, &4=time selection,
                &8=all project regions, &16=selected media items,
                &32=selected project regions
            RENDER_TAILMS : tail length in ms to render
                (only used if RENDER_BOUNDSFLAG and RENDER_TAILFLAG are set)
            RENDER_ADDTOPROJ : 1=add rendered files to project
            RENDER_DITHER : &1=dither, &2=noise shaping, &4=dither stems,
                &8=noise shaping on stems
            PROJECT_SRATE : samplerate (ignored unless PROJECT_SRATE_USE set)
            PROJECT_SRATE_USE : set to 1 if project samplerate is used
        param_value : float
        """
        RPR.GetSetProjectInfo(  # type:ignore
            self.id, param_name, param_value, True)

    def set_ext_state(
        self,
        section: str,
        key: str,
        value: Union[str, object],
        pickled: bool = False
    ) -> None:
        """
        Set external state of project.

        Parameters
        ----------
        section : str
        key : str
        value : Union[Any, str]
            State value. Will be dumped to str using either `pickle` if
            `pickled` is `True` or `json`. Length of the dumped value
            must not be over 2**31 - 2.
        pickled : bool, optional
            Data will be pickled with the last version if True.
            If you using mypy as type checker, typing_extensions.Literal[True]
            has to be used for `pickled`.

        Raises
        ------
        ValueError
            If dumped `value` has length over 2**31 - 2.
        """
        if pickled:
            value = pickle.dumps(value)
            value = codecs.encode(value, "base64").decode()
        if not isinstance(value, (str, bytes)):
            raise TypeError(
                "value has to be of type 'str', or should be picked"
            )
        if len(value) > 2**31 - 2:
            message = (
                "Dumped value length is {:,d}. It must not be over "
                "2**31 - 2."
            )
            raise ValueError(message.format(len(value)))
        RPR.SetProjExtState(self.id, section, key, value)  # type:ignore

    @reapy_boost.inside_reaper()
    def solo_all_tracks(self) -> None:
        """
        Solo all tracks in project.

        See also
        --------
        Project.unsolo_all_tracks
        """
        with self.make_current_project():
            RPR.SoloAllTracks(1)  # type:ignore

    def stop(self) -> None:
        """
        Hit stop button.
        """
        RPR.OnStopButtonEx(self.id)  # type:ignore

    @property
    def time_selection(self) -> 'reapy_boost.TimeSelection':
        """
        Project time selection.

        time_selection : reapy_boost.TimeSelection

        Can be set and deleted as follows:

        >>> project = reapy_boost.Project()
        >>> project.time_selection = 3, 8  # seconds
        >>> del project.time_selection
        """
        return self._time_selection_inside()

    @time_selection.setter
    def time_selection(self, selection: Tuple[float, float]) -> None:
        """
        Set time selection bounds.

        Parameters
        ----------
        selection : (float, float)
            Start and end of new time selection in seconds.
        """
        self.time_selection._set_start_end(*selection)

    @time_selection.deleter
    def time_selection(self) -> None:
        """
        Delete current time selection.
        """
        self.time_selection._set_start_end(0, 0)

    @reapy_boost.inside_reaper()
    def _time_selection_inside(self) -> 'reapy_boost.TimeSelection':
        time_selection = reapy_boost.TimeSelection(self)
        return time_selection

    @property
    def time_signature(self) -> Tuple[float, int]:
        """
        Project time signature.

        This does not reflect tempo envelopes but is purely what is set in the
        project settings.

        bpm : float
            Project BPM (beats per minute)
        bpi : int
            Project BPI (numerator of time signature)
        """
        _, bpm, bpi = RPR.GetProjectTimeSignature2(  # type:ignore
            self.id, 0, 0)
        return bpm, int(bpi)

    def time_signature_at_position(
        self, time: float
    ) -> Tuple[float, Tuple[int, int]]:
        """Tempo and time signature at the time.time_selection

        Parameters
        ----------
        time : float
            in seconds

        Returns
        -------
        Tuple[float, Tuple[int,int]]
            tempo (bpm): float
            numerator, denomimator Tuple[int,int]
        """
        (_, _, num, denom,
         tempo) = RPR.TimeMap_GetTimeSigAtTime(  # type:ignore
         self.id, time, 0, 0, 0)
        return tempo, (int(num), int(denom))

    def time_to_beats(self, time: float) -> float:
        """
        Convert time in seconds to beats.

        Parameters
        ----------
        time : float
            Time in seconds.

        Returns
        -------
        beats : float
            Time in beats.

        See also
        --------
        Projecr.beats_to_time
        """
        beats = RPR.TimeMap2_timeToQN(self.id, time)  # type:ignore
        return cast(float, beats)

    @property
    def tracks(self) -> 'reapy_boost.TrackList':
        """
        List of project tracks.

        :type: TrackList
        """
        return reapy_boost.TrackList(self)

    def undo(self) -> None:
        """
        Undo last action.

        Raises
        ------
        UndoError
            If impossible to undo.
        """
        success = RPR.Undo_DoUndo2(self.id)  # type:ignore
        if not success:
            raise UndoError

    def unmute_all_tracks(self) -> None:
        """
        Unmute all tracks.
        """
        self.mute_all_tracks(mute=False)

    def unselect_all_tracks(self) -> None:
        """Unselect all tracks."""
        self.perform_action(40297)

    @reapy_boost.inside_reaper()
    def unsolo_all_tracks(self) -> None:
        """
        Unsolo all tracks in project.

        See also
        --------
        Project.solo_all_tracks
        """
        with self.make_current_project():
            RPR.SoloAllTracks(0)  # type:ignore


class _MakeCurrentProject:
    """Context manager used by Project.make_current_project."""

    def __init__(self, project: Project):
        self.current_project = self._make_current_project(project)

    @staticmethod
    @reapy_boost.inside_reaper()
    def _make_current_project(project: Project) -> Project:
        """Set current project and return the previous current project."""
        current_project = reapy_boost.Project()
        RPR.SelectProjectInstance(project.id)  #type:ignore
        return current_project

    def __enter__(self) -> None:
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):  #type:ignore
        # Test for valid ID in case project has been closed since __enter__
        if self.current_project.has_valid_id:
            self.current_project.make_current_project()



"""Define tools such as Program and custom json module."""

import reapy_boost
from ._inside_reaper import inside_reaper, dist_api_is_enabled
from .network.machines import connect, connect_to_default_machine, reconnect
from .extension_dependency import (
    depends_on_sws, depends_on_extension, generate_imgui
)

__all__ = [
    "inside_reaper",
    "dist_api_is_enabled",
    "connect",
    "connect_to_default_machine",
    "reconnect",
    "depends_on_sws",
    "depends_on_extension",
    "generate_imgui",
]


from typing import Dict, Optional
from reapy_boost import reascript_api as RPR
from reapy_boost.core import ReapyObject
import reapy_boost


class AutomationItem(ReapyObject):

    _class_name = "AutomationItem"

    def __init__(
        self,
        envelope: Optional['reapy_boost.Envelope'] = None,
        index: int = 0,
        envelope_id: int = None
    ) -> None:
        if envelope is not None:
            envelope_id = envelope.id
        if envelope_id is None:
            raise TypeError(
                'Either envelope or envelope_id has to be specified'
            )
        self.envelope_id = envelope_id
        self.index = index

    @property
    def _kwargs(self) -> Dict[str, int]:
        return {"index": self.index, "envelope_id": self.envelope_id}

    def delete_points_in_range(self, start: float, end: float) -> None:
        """
        Delete points between `start` and `end`.

        Parameters
        ----------
        start : float
            Range start in seconds.
        end : float
            Range end in seconds.
        """
        RPR.DeleteEnvelopePointRangeEx(  # type:ignore
            self.envelope_id, self.index, start, end
        )

    @property
    def length(self) -> float:
        """
        Return item length in seconds.

        Returns
        -------
        length : float
            Item length in seconds.
        """
        length = RPR.GetSetAutomationItemInfo(  # type:ignore
            self.envelope_id, self.index, "D_LENGTH", 0, False
        )
        return length

    @length.setter
    def length(self, length: float) -> None:
        """
        Set item length.

        Parameters
        ----------
        length : float
            New item length in seconds.
        """
        success = RPR.GetSetAutomationItemInfo(  # type:ignore
            self.envelope_id, self.index, "D_LENGTH", length, True
        )

    @property
    def n_points(self) -> int:
        """
        Return number of automation points in item.

        Returns
        -------
        n_points : int
            Number of automation points in item.
        """
        n_points = RPR.CountEnvelopePointsEx(  # type:ignore
            self.envelope_id, self.index)
        return n_points

    @property
    def pool(self) -> int:
        """
        Return item pool.

        Returns
        -------
        pool : int
            Item pool.
        """
        pool = RPR.GetSetAutomationItemInfo(  # type:ignore
            self.envelope_id, self.index, "D_POOL", 0, False
        )
        return pool

    @pool.setter
    def pool(self, pool: int) -> None:
        """
        Set item pool.

        Parameters
        ----------
        pool : int
            New item pool.
        """
        success = RPR.GetSetAutomationItemInfo(  # type:ignore
            self.envelope_id, self.index, "D_POOL", pool, True
        )

    @property
    def position(self) -> float:
        """
        Return item position in seconds.

        Returns
        -------
        position : float
            Item position in seconds.
        """
        position = RPR.GetSetAutomationItemInfo(  # type:ignore
            self.envelope_id, self.index, "D_POSITION", 0, False
        )
        return position

    @position.setter
    def position(self, position: float) -> None:
        """
        Set item position.

        Parameters
        ----------
        position : float
            New item position in seconds.
        """
        success = RPR.GetSetAutomationItemInfo(  # type:ignore
            self.envelope_id, self.index, "D_POSITION", position, True
        )



"""
Disable ```reapy_boost`` distant API.

Running this ReaScript from inside REAPER disables ``reapy_boost`` imports
from outside. It deletes the persistent Web Interface and removes the
ReaScript ``reapy_boost.reascripts.activate_reapy_server`` from the Actions
list.

See also
--------
reapy_boost.reascripts.enable_dist_api
"""

if __name__ == "__main__":
    import reapy_boost
    reapy_boost.config.disable_dist_api()



from typing import Dict, Optional, TypedDict, Union, cast
import reapy_boost
from reapy_boost import reascript_api as RPR
from reapy_boost.core import ReapyObject


class MarkerInfo(TypedDict):

    index: int
    enum_index: int
    project_id: str
    position: float
    name: str


class Marker(ReapyObject):

    _class_name = "Marker"

    def __init__(
        self,
        parent_project: Optional['reapy_boost.Project'] = None,
        index: Optional[int] = None,
        parent_project_id: Optional[str] = None,
        enum_index: Optional[int] = None
    ):
        if parent_project_id is None:
            message = (
                "One of `parent_project` or `parent_project_id` must be " +
                "specified."
            )
            assert parent_project is not None, message
            parent_project_id = parent_project.id
        self.project = reapy_boost.Project(parent_project_id)
        self.project_id = parent_project_id
        if index is None:
            index = self.project.n_markers
        self.index = index
        if enum_index is None:
            self.enum_index = None
            enum_index = self._get_enum_index()
        self.enum_index = enum_index

    def _get_enum_index(self) -> int:
        """
        Return marker index as needed by RPR.EnumProjectMarkers2.

        Raises
        ------
        reapy_boost.errors.UndefinedMarkerError
            Description
        """
        with reapy_boost.inside_reaper():
            for marker in self.project.markers:
                if marker.index == self.index:
                    return marker.enum_index
        raise reapy_boost.errors.UndefinedMarkerError(self.index)

    @property
    def _kwargs(self) -> Dict[str, object]:
        return {
            "index": self.index,
            "parent_project_id": self.project_id,
            'enum_index': self.enum_index
        }

    def delete(self) -> None:
        """
        Delete marker.
        """
        RPR.DeleteProjectMarker(  # type:ignore
            self.project_id, self.index, False)

    @property
    def infos(self) -> MarkerInfo:
        """Get all Region infos in one call.

        Returns
        -------
        MarkerInfo
            index: int
            enum_index: int
            project_id: str
            position: float
            name: str
        """
        return self._infos_inside()

    @reapy_boost.inside_reaper()
    def _infos_inside(self) -> 'MarkerInfo':
        enum_index = self._get_enum_index()
        args = self.project_id, enum_index, 0, 0, 0, 0, 0
        (_, _, _, _, position, _, _,
         index) = RPR.EnumProjectMarkers2(  # type:ignore
             *args
         )
        out: MarkerInfo = {
            'index': cast(int, index),
            'enum_index': enum_index,
            'project_id': self.project_id,
            'position': cast(float, position),
            'name': self.name,
        }
        return out

    @property
    def name(self) -> str:
        """
        Marker name.

        :type: str
        """
        with reapy_boost.inside_reaper():
            # index = self._get_enum_index()
            fs = RPR.SNM_CreateFastString('0' * 1024)  # type:ignore
            args = self.project_id, self.index, False, fs
            RPR.SNM_GetProjectMarkerName(*args)  # type:ignore
            result = RPR.SNM_GetFastString(fs)  # type:ignore
            RPR.SNM_DeleteFastString(fs)  # type:ignore
        return result

    @name.setter
    def name(self, name: str) -> None:
        """
        Set marker name.

        Parameters
        ----------
        name : str
        """
        args = self.project_id, self.index, False, self.position, 0, name
        RPR.SetProjectMarker2(*args)  # type:ignore

    @property
    def position(self) -> float:
        """
        Return marker position.

        Returns
        -------
        position : float
            Marker position in seconds.
        """
        return self._position_inside()

    @reapy_boost.inside_reaper()
    def _position_inside(self) -> float:
        index = self._get_enum_index()
        print(f'enum index in position: {index}')
        return RPR.EnumProjectMarkers2(  # type:ignore
            self.project_id, index, 0, 0, 0, 0, 0)[4]

    @position.setter
    def position(self, position: float) -> None:
        """
        Set marker position.

        Parameters
        ----------
        position : float
            Marker position in seconds.
        """
        RPR.SetProjectMarker2(  # type:ignore
            self.project_id, self.index, False, position, 0, self.name)



"""API for ReaExtensions by Julian Sader.

Repo on GitHub: https://github.com/juliansader/ReaExtensions
Theme on ReaperForums: https://forum.cockos.com/showthread.php?t=212174

Module consists of autogenerated members and some "fixed" functions, which
ought to replace autogenerated.
But the "fixed" functions are renamed for avoiding umbiguos behaviour.
For example: instead of MIDIEditor_ArrayAll use MIDIEditor_ArrayAllEx.
"""

import typing as ty
import reapy_boost
import ctypes as ct
if reapy_boost.is_inside_reaper():
    from reapy_boost.additional_api import packp, unpackp, packs_l, unpacks_l
    from reaper_python import _ft
try:
    from ._JS_API_generated import *
    from ._JS_API_generated import VoidPtr
except ImportError:
    from warnings import warn
    warn(
        ImportWarning(
            'JS API still is not generated, or generated incorrectly.'
        )
    )

# initial __all__ is autogenerated, be careful!
__all__: ty.List[str] = [
    "Pointer", "VoidPtr", "LICE_IBitmap", "PCM_source", "AudioWriter",
    "ReaScriptAPI_Version", "Localize", "Mem_Alloc", "Mem_Free",
    "Mem_FromString", "String", "Int", "Byte", "Double",
    "Dialog_BrowseForSaveFile", "Dialog_BrowseForOpenFiles",
    "Dialog_BrowseForFolder", "Window_GetRect", "Window_ScreenToClient",
    "Window_ClientToScreen", "Window_GetClientRect", "Window_GetClientSize",
    "Window_MonitorFromRect", "Window_GetViewportFromRect", "Window_Update",
    "Window_InvalidateRect", "Window_FromPoint", "Window_GetParent",
    "Window_SetParent", "Window_IsChild", "Window_GetRelated",
    "Window_FindChildByID", "Window_SetFocus", "Window_GetFocus",
    "Window_SetForeground", "Window_GetForeground", "Window_Create",
    "Window_EnableMetal", "Window_Enable", "Window_Destroy", "Window_Show",
    "Window_IsVisible", "Window_IsWindow", "Window_FindEx", "Window_Find",
    "Window_FindTop", "Window_FindChild", "Window_ArrayAllChild",
    "Window_ArrayAllTop", "Window_ArrayFind", "Window_ListAllChild",
    "Window_ListAllTop", "Window_ListFind", "MIDIEditor_ListAll",
    "MIDIEditor_ArrayAll", "Window_SetPosition", "Window_Resize",
    "Window_Move", "Window_SetZOrder", "Window_GetLongPtr", "Window_GetLong",
    "Window_SetLong", "Window_SetStyle", "Window_SetOpacity",
    "Window_GetTitle", "Window_SetTitle", "Window_GetClassName",
    "Window_HandleFromAddress", "Window_AddressFromHandle",
    "WindowMessage_Post", "WindowMessage_Send", "WindowMessage_Peek",
    "WindowMessage_Intercept", "WindowMessage_InterceptList",
    "WindowMessage_PassThrough", "WindowMessage_ListIntercepts",
    "WindowMessage_Release", "WindowMessage_ReleaseWindow",
    "WindowMessage_ReleaseAll", "Window_OnCommand", "VKeys_GetState",
    "VKeys_GetDown", "VKeys_GetUp", "VKeys_Intercept", "Mouse_GetState",
    "Mouse_GetCursor", "Mouse_SetPosition", "Mouse_LoadCursor",
    "Mouse_LoadCursorFromFile", "Mouse_SetCursor", "Window_GetScrollInfo",
    "Window_SetScrollPos", "GDI_GetClientDC", "GDI_GetWindowDC",
    "GDI_GetScreenDC", "GDI_ReleaseDC", "GDI_CreateFillBrush", "GDI_CreatePen",
    "GDI_CreateFont", "GDI_SelectObject", "GDI_DeleteObject", "GDI_FillRect",
    "GDI_FillRoundRect", "GDI_FillPolygon", "GDI_FillEllipse",
    "GDI_GetSysColor", "GDI_SetTextBkMode", "GDI_SetTextBkColor",
    "GDI_SetTextColor", "GDI_GetTextColor", "GDI_DrawText", "GDI_SetPixel",
    "GDI_Line", "GDI_Polyline", "GDI_Blit", "GDI_StretchBlit", "Composite",
    "Composite_Unlink", "Composite_ListBitmaps", "Composite_Delay",
    "LICE_CreateBitmap", "LICE_ListAllBitmaps", "LICE_ArrayAllBitmaps",
    "LICE_GetHeight", "LICE_GetWidth", "LICE_GetDC", "LICE_DestroyBitmap",
    "LICE_LoadPNG", "LICE_Blit", "LICE_RotatedBlit", "LICE_ScaledBlit",
    "LICE_IsFlipped", "LICE_Resize", "LICE_Clear", "LICE_CreateFont",
    "LICE_DestroyFont", "LICE_SetFontFromGDI", "LICE_SetFontColor",
    "LICE_SetFontBkColor", "LICE_DrawText", "LICE_DrawChar",
    "LICE_MeasureText", "LICE_GradRect", "LICE_FillRect", "LICE_FillTriangle",
    "LICE_FillPolygon", "LICE_FillCircle", "LICE_Line", "LICE_Bezier",
    "LICE_Arc", "LICE_Circle", "LICE_RoundRect", "LICE_GetPixel",
    "LICE_PutPixel", "LICE_WritePNG", "LICE_WriteJPG", "LICE_LoadJPG",
    "LICE_SetAlphaFromColorMask", "LICE_AlterBitmapHSV", "LICE_AlterRectHSV",
    "LICE_ProcessRect", "Window_AttachTopmostPin", "Window_AttachResizeGrip",
    "ListView_GetItemCount", "ListView_GetSelectedCount",
    "ListView_EnsureVisible", "ListView_GetFocusedItem",
    "ListView_EnumSelItems", "ListView_GetItem", "ListView_GetItemText",
    "ListView_GetItemState", "ListView_ListAllSelItems",
    "Xen_AudioWriter_Create", "Xen_AudioWriter_Destroy",
    "Xen_AudioWriter_Write", "Xen_GetMediaSourceSamples",
    "Xen_StartSourcePreview", "Xen_StopSourcePreview"
]
# This part is where the module __all__ begins
__all__.extend(
    [
        "Window_ArrayAllChildEx",
        "Window_ArrayAllTopEx",
        "Window_ArrayFindEx",
        "MIDIEditor_ArrayAllEx",
        "reaper_array_to_hwnd",
    ]
)


@reapy_boost.inside_reaper()
def create_reaper_array(size: int = 100) -> ct.Structure:

    class ReaperArray(ct.Structure):
        _fields_ = [
            ('size', ct.c_uint32), ('maximum', ct.c_uint32),
            ('values', ct.c_double * size)
        ]

    return ReaperArray(size=0, maximum=size, values=(ct.c_double * size)(0.0))


@reapy_boost.inside_reaper()
def Window_ArrayAllChildEx(
    parentHWND: 'VoidPtr',
    size: int = 100
) -> ty.Tuple[int, ty.Tuple[float, ...]]:
    """
    Finds all child windows of the specified parent.

    Returns:
     * retval:
    The number of windows found; negative if an error occurred.
     * The
    addresses are stored in the provided reaper.array, and can be
    converted to REAPER objects (HWNDs) by the function
    JS_Window_HandleFromAddress.

    Returns
    -------
    ret_value: int
    array: Tuple[float]
    """
    a = _ft['JS_Window_ArrayAllChild']
    f = ct.CFUNCTYPE(ct.c_int, ct.c_void_p, ct.c_void_p)(a)
    array = create_reaper_array(size)
    ret = f(packp("void*", str(parentHWND)), ct.byref(array))
    return int(ret), tuple(array.values)


@reapy_boost.inside_reaper()
def Window_ArrayAllTopEx(
    size: int = 100
) -> ty.Tuple[int, ty.Tuple[float, ...]]:
    """
    Finds all top-level windows.

    Returns:
     * retval: The number of
    windows found; negative if an error occurred.
     * The addresses are
    stored in the provided reaper.array, and can be converted to REAPER
    objects (HWNDs) by the function JS_Window_HandleFromAddress.

    Returns
    -------
    ret_value: int
    array: Tuple[float]
    """
    a = _ft['JS_Window_ArrayAllTop']
    f = ct.CFUNCTYPE(ct.c_int, ct.c_void_p)(a)
    array = create_reaper_array(size)
    ret = f(ct.byref(array))
    return int(ret), tuple(array.values)


@reapy_boost.inside_reaper()
def Window_ArrayFindEx(
    title: str,
    exact: bool,
    size: int = 100,
    encoding: str = "utf-8"
) -> ty.Tuple[int, ty.Tuple[float, ...]]:
    """
    Finds all windows, whether top-level or child, whose titles match the
    specified string.

    Returns:
     * retval: The number of windows found;
    negative if an error occurred.
     * The addresses are stored in the
    provided reaper.array, and can be converted to REAPER objects (HWNDs)
    by the function JS_Window_HandleFromAddress.

    Parameters:
     * exact:
    Match entire title exactly, or match substring of title.

    Returns
    -------
    ret_value: int
    array: Tuple[float]
    """
    a = _ft['JS_Window_ArrayFind']
    f = ct.CFUNCTYPE(ct.c_int, ct.c_char_p, ct.c_byte, ct.c_void_p)(a)
    prot_title = packs_l(title, size=len(title), encoding=encoding)
    array = create_reaper_array(size)
    ret = f(prot_title, ct.c_byte(exact), ct.byref(array))
    return int(ret), tuple(array.values)


@reapy_boost.inside_reaper()
def MIDIEditor_ArrayAllEx(
    size: int = 100,
) -> ty.Tuple[int, ty.Tuple[float, ...]]:
    """
    Finds all open MIDI windows (whether docked or not).

     * retval: The
    number of MIDI editor windows found; negative if an error occurred.
    * The address of each MIDI editor window is stored in the provided
    reaper.array. Each address can be converted to a REAPER object (HWND)
    by the function JS_Window_HandleFromAddress.

    Returns
    -------
    ret_value: int
    array: Tuple[float]
    """
    a = _ft['JS_MIDIEditor_ArrayAll']
    f = ct.CFUNCTYPE(ct.c_int, ct.c_void_p)(a)
    array = create_reaper_array(size)
    ret = f(ct.byref(array))
    return int(ret), tuple(array.values)


@reapy_boost.inside_reaper()
def LICE_LoadPNGEx(filename: str, encoding: str = "utf-8") -> 'VoidPtr':
    """
    Returns a system LICE bitmap containing the PNG.

    Note
    ----
    This is stable version of LoadPNG which provides constant filename
    string size of 2048. God knows, why, but we respect him.

    Returns
    -------
    ret_value: VoidPtr
    """
    a = _ft['JS_LICE_LoadPNG']
    f = ct.CFUNCTYPE(ct.c_void_p, ct.c_char_p)(a)
    prot_filename = packs_l(
        filename,
        # size=len(filename.encode(encoding)),
        size=2048,  # strange need to keep it constant
        encoding=encoding,
    )
    ret = f(prot_filename)
    return VoidPtr(unpackp("void*", ret))


@reapy_boost.inside_reaper()
def reaper_array_to_hwnd(array: ty.Tuple[float, ...]) -> ty.List['VoidPtr']:
    """Transform reaper array to VoidPtr pointers to hwnd.

    Parameters
    ----------
    array : ty.Tuple[float, ...]
        got within *Array funtions

    Returns
    -------
    ty.List[VoidPtr]
    """
    out = list()
    for val in array:
        if val != 0:
            out.append(VoidPtr(int(val)))
    return out



import typing as ty

import reapy_boost
from reapy_boost import reascript_api as RPR
from reapy_boost.core.item.take import Take
from reapy_boost.core.reapy_object import ReapyObject
from reapy_boost.core.track.automation_item import AutomationItem
from reapy_boost.core.track.track import Track


class EnvelopePoint(ty.TypedDict):
    """Dictionary represents envelope point.

    Members
    -------
    index: int
    time: float
    value: float
    shape: int
    tension: float
    selected: float
    """

    index: int
    time: float
    value: float
    shape: int
    tension: float
    selected: float


class Envelope(ReapyObject):

    def __init__(self, parent: ty.Union[Track, Take], id: str) -> None:
        self.id = id
        self._parent = parent

    @property
    def _args(self) -> ty.Tuple[ty.Union[Track, Take], str]:
        return (self.parent, self.id)

    def add_item(
        self,
        position: float = 0.,
        length: float = 1.,
        pool: int = 0
    ) -> AutomationItem:
        """
        Add automation item to envelope.

        Parameters
        ----------
        position : float, optional
            New item position in seconds (default=0).
        length : float
            New item length in seconds (default=0).
        pool : int
            New item pool index. If >= 0 the automation item will be a
            new instance of that pool (which will be created as an
            empty instance if it does not exist).

        Returns
        -------
        item : reapy_boost.AutomationItem
            New automation item.
        """
        item_index = RPR.InsertAutomationItem(  # type: ignore
            self.id,
            pool,
            position,
            length
        )
        item = reapy_boost.AutomationItem(envelope=self, index=item_index)
        return item

    def delete_points_in_range(self, start: float, end: float) -> None:
        """
        Delete envelope points between `start` and `end`.

        Parameters
        ----------
        start : float
            Range start in seconds.
        end : float
            Range end in seconds.
        """
        RPR.DeleteEnvelopePointRange(self.id, start, end)  # type:ignore

    @reapy_boost.inside_reaper()
    def get_derivatives(self,
                        time: float,
                        raw: bool = False) -> ty.Tuple[float, float, float]:
        """
        Return envelope derivatives of order 1, 2, 3 at a given time.

        Parameters
        ----------
        time : float
            Time in seconds.
        raw : bool, optional
            Whether to return raw values or the human-readable version
            which is printed in REAPER GUI (default=False).

        Returns
        -------
        d, d2, d3 : float
            First, second and third order derivatives.

        Examples
        --------
        >>> envelope = track.envelopes["Pan"]
        >>> envelope.get_derivatives(10, raw=True)
        (0.10635556358181712, 0.2127113398749741, 0.21271155258652666)
        >>> envelope.get_value(10)  # human-readable
        ('10%L', '21%L', '21%L')
        """
        d, d2, d3 = RPR.Envelope_Evaluate(  # type:ignore
            self.id,
            time,
            1,
            1,
            0,
            0,
            0,
            0
        )[6:]
        if not raw:
            d = RPR.Envelope_FormatValue(  # type:ignore
                self.id, d, "", 2048
            )[2]
            d2 = RPR.Envelope_FormatValue(  # type:ignore
                self.id,
                d2,
                "",
                2048
            )[2]
            d3 = RPR.Envelope_FormatValue(  # type:ignore
                self.id,
                d3,
                "",
                2048
            )[2]
        return d, d2, d3

    @reapy_boost.inside_reaper()
    def get_value(self,
                  time: float,
                  raw: bool = False) -> ty.Union[float, str]:
        """
        Return envelope value at a given time.

        Parameters
        ----------
        time : float
            Time in seconds.
        raw : bool, optional
            Whether to return raw value or its human-readable version,
            which is the one that is printed in REAPER GUI
            (default=False).

        Returns
        -------
        value : float or str
            Envelope value.

        Examples
        --------
        >>> envelope = track.envelopes["Pan"]
        >>> envelope.get_value(10, raw=True)
        -0.5145481809245827
        >>> envelope.get_value(10)  # human-readable
        '51%R'
        """
        value = RPR.Envelope_Evaluate(  # type:ignore
            self.id, time, 0, 0, 0, 0, 0, 0)[5]
        if not raw:
            value = RPR.Envelope_FormatValue(  # type:ignore
                self.id, value, "", 2048)[2]
        return value

    def get_point(self, index: int) -> EnvelopePoint:
        """Get EvnelopePoint as dictionary.

        Parameters
        ----------
        index : int
            point index

        Returns
        -------
        EnvelopePoint
        """
        (_, _, index, time, value, shape, tension,
         selected) = RPR.GetEnvelopePoint(  # type:ignore
             self.id,
             index,
             0.0,
             0.0,
             0,
             0.0,
             1
         )
        return EnvelopePoint(
            index=index,
            time=time,
            value=value,
            shape=shape,
            tension=tension,
            selected=selected,
        )

    @property
    def has_valid_id(self) -> bool:
        """
        Whether ReaScript ID is still valid.

        For instance, if envelope has been deleted, ID will not be valid
        anymore.

        :type: bool
        """
        return self._has_valid_id_inside()

    @reapy_boost.inside_reaper()
    def _has_valid_id_inside(self) -> bool:
        try:
            project_id = self.parent.project.id
        except (OSError, AttributeError):
            return False
        pointer, name = self._get_pointer_and_name()
        return bool(
            RPR.ValidatePtr2(  # type:ignore
                project_id, pointer, name
            )
        )

    def insert_point(self, point: EnvelopePoint, sort: bool = True) -> bool:
        """Insert EnvelopePoint

        Parameters
        ----------
        point : EnvelopePoint
            index not counted
        sort : bool, optional
            set to False if insert many points,
            than use self.sort_points()

        Returns
        -------
        bool
        """
        (retval, _, _, _, _, _, _, _) = RPR.InsertEnvelopePoint(  # type:ignore
            self.id, point['time'], point['value'], point['shape'],
            point['tension'], point['selected'], not sort
        )
        return retval

    @property
    def items(self) -> ty.List[AutomationItem]:
        """
        List of automation items in envelope.

        :type: list of reapy_boost.AutomationItem
        """
        items = [
            reapy_boost.AutomationItem(self, i) for i in range(self.n_items)
        ]
        return items

    @property
    def n_items(self) -> int:
        """
        Number of automation items in envelope.

        :type: int
        """
        n_items = RPR.CountAutomationItems(self.id)  # type:ignore
        return n_items

    @property
    def n_points(self) -> int:
        """
        Number of points in envelope.

        :type: int
        """
        n_points = RPR.CountEnvelopePoints(self.id)  # type:ignore
        return n_points

    @property
    def name(self) -> str:
        """
        Envelope name.

        :type: str
        """
        name = RPR.GetEnvelopeName(self.id, "", 2048)[2]  # type:ignore
        return name

    @property
    def parent(self) -> ty.Union[Track, Take]:
        """
        Envelope parent.

        :type: Take or Track
        """
        return self._parent

    def set_point(
        self, index: int, value: EnvelopePoint, sort: bool = True
    ) -> bool:
        """Set envelope point.

        Note
        ----
        Not tested yet

        Parameters
        ----------
        index : int
        value : EnvelopePoint
        sort : bool, optional
            set to False if set many points,
            than use self.sort_points()
        """
        (retval, _, _, _, _, _, _, _, _) = RPR.SetEnvelopePoint(  # type:ignore
            self.id, index, value['time'], value['value'], value['shape'],
            value['tension'], value['selected'], not sort
        )
        return bool(retval)

    def sort_points(self) -> None:
        RPR.Envelope_SortPoints(self.id)  # type:ignore


class EnvelopeList(ReapyObject):
    """
    Container class for the list of envelopes on a Take or Track.

    Envelopes can be accessed from the EnvelopeList either by index,
    name or chunk_name (e.g. "<VOLENV").

    Examples
    --------
    >>> len(track.envelopes)
    2
    >>> envelope = track.envelopes[0]
    >>> envelope.name
    'Volume'
    >>> envelope == track.envelopes["Volume"]
    True
    >>> envelope == track.envelopes["<VOLENV"]
    True
    >>> [e.name for e in track.envelopes]
    ['Volume', 'Pan']
    """

    def __init__(self, parent: ty.Union[Track, Take]):
        self.parent = parent

    @property
    def _args(self) -> ty.Tuple[ty.Union[Track, Take]]:
        return (self.parent, )

    def __getitem__(self, key: ty.Union[str, int]) -> Envelope:
        parent_type = self.parent.__class__._reapy_parent.__name__
        attr = "Get{}Envelope".format(parent_type)
        if isinstance(key, str):
            if key.startswith("<") and parent_type == 'Track':
                attr += "ByChunkName"
            else:
                attr += "ByName"
        callback = getattr(RPR, attr)
        envelope = Envelope(self.parent, callback(self.parent.id, key))
        if not envelope._is_defined:
            raise KeyError("No envelope for key {}".format(repr(key)))
        return envelope

    def __len__(self) -> int:
        return self.parent.n_envelopes



"""Define Server class."""

import reapy_boost
from reapy_boost.tools import json
from .socket import Socket

import socket
import traceback
import typing as ty


class Server(Socket):
    """
    Server part of the ``reapy_boost`` dist API.

    It is instantiated inside REAPER. It receives and processes API
    call requests coming from the outside.
    """
    connections: ty.Dict[ty.Union[ty.Tuple[str, ...], str], Socket]

    def __init__(self, port: int) -> None:
        ...

    @Socket._non_blocking
    def _get_request(
        self, connection: Socket, address: ty.Union[ty.Tuple[str, ...], str]
    ) -> ty.Dict[str, object]:
        ...

    def _hold_connection(
        self, address: ty.Union[ty.Tuple[str, ...], str]
    ) -> ty.Dict[str, ty.Optional[str]]:
        ...

    def _process_request(
        self, request: ty.Dict[str, object],
        address: ty.Union[ty.Tuple[str, ...], str]
    ) -> ty.Dict[str, ty.Any]:
        ...

    def _send_result(
        self, connection: Socket, result: ty.Dict[str, ty.Any]
    ) -> None:
        ...

    @Socket._non_blocking
    def accept(
        self, *args: ty.Any, **kwargs: ty.Any
    ) -> ty.Tuple['Socket', ty.Union[str, ty.Tuple[str, ...]]]:
        ...

    def disconnect(self, address: ty.Union[ty.Tuple[str, ...], str]) -> None:
        ...

    def get_requests(self) -> ty.Dict[str, ty.Any]:
        ...

    def process_requests(
        self, requests: ty.Dict[str, ty.Any]
    ) -> ty.Dict[str, ty.Any]:
        ...

    def send_results(self, results: ty.Dict[str, ty.Any]) -> None:
        ...



"""Audio handling functions."""

from enum import Enum
from typing import List, Tuple
import reapy_boost
import reapy_boost.reascript_api as RPR


def get_input_latency(unit: str = "second") -> float:
    """
    Return input latency.

    Parameters
    ----------
    unit : {"sample", "second"}
        Whether to return latency in samples or seconds
        (default="second").

    Returns
    -------
    latency : float
        Input latency.
    """
    latency, out_latency = RPR.GetInputOutputLatency(0, 0)  # type:ignore
    if unit == "second":
        # Small hack because RPR.GetInputLatency doesn't exist...
        latency *= RPR.GetOutputLatency() / out_latency  # type:ignore
    return latency


@reapy_boost.inside_reaper()
def get_input_names() -> Tuple[str, ...]:
    """
    Return names of all input channels.

    Returns
    -------
    names : list of str
        Names of input channels.
    """
    n_channels = reapy_boost.audio.get_n_inputs()
    return tuple(
        map(
            RPR.GetInputChannelName,  # type:ignore
            range(n_channels)
        )
    )


def get_n_inputs() -> int:
    """
    Return number of audio inputs.

    Returns
    -------
    n_inputs : int
        Number of audio inputs.
    """
    n_inputs = RPR.GetNumAudioInputs()  # type:ignore
    return n_inputs


def get_n_outputs() -> int:
    """
    Return number of audio outputs.

    Returns
    -------
    n_outputs : int
        Number of audio outputs.
    """
    n_outputs = RPR.GetNumAudioOutputs()  # type:ignore
    return n_outputs


def get_output_latency(unit: str = "second") -> float:
    """
    Return output latency.

    Parameters
    ----------
    unit : {"sample", "second"}
        Whether to return latency in samples or seconds
        (default="second").

    Returns
    -------
    latency : float
        Output latency.
    """
    latency, out_latency = RPR.GetInputOutputLatency(0, 0)  # type:ignore
    if unit == "second":
        latency = RPR.GetOutputLatency()  # type:ignore
    else:
        latency = RPR.GetInputOutputLatency(0, 0)[1]  # type:ignore
    return latency


@reapy_boost.inside_reaper()
def get_output_names() -> Tuple[str, ...]:
    """
    Return names of all output channels.

    Returns
    -------
    names : list of str
        Names of output channels.
    """
    n_channels = reapy_boost.audio.get_n_outputs()
    return tuple(
        map(
            RPR.GetOutputChannelName,  # type:ignore
            range(n_channels)
        )
    )


def init() -> None:
    """
    Open all audio and MIDI devices (if not opened).
    """
    RPR.Audio_Init()  # type:ignore


def is_prebuffer() -> bool:
    """
    Return whether audio is in pre-buffer (threadsafe).

    Returns
    -------
    is_prebuffer : bool
        Whether audio is in pre-buffer.
    """
    is_prebuffer = bool(RPR.Audio_IsPreBuffer())  # type:ignore
    return is_prebuffer


def is_running() -> bool:
    """
    Return whether audio is running (threadsafe).

    Returns
    -------
    is_running : bool
        Whether audio is running.
    """
    is_running = bool(RPR.Audio_IsRunning())  # type:ignore
    return is_running


def quit() -> None:
    """
    Close all audio and MIDI devices (if opened).
    """
    RPR.Audio_Quit()  # type:ignore



from . import audio, midi, ui
from .reaper import *
from .defer import defer, at_exit

__all__ = [
    'add_reascript',
    'arm_command',
    'browse_for_file',
    'clear_console',
    'clear_peak_cache',
    'dB_to_slider',
    'delete_ext_state',
    'disarm_command',
    'get_armed_command',
    'get_command_id',
    'get_command_name',
    'get_exe_dir',
    'get_ext_state',
    'get_global_automation_mode',
    'get_ini_file',
    'get_last_touched_track',
    'get_main_window',
    'get_projects',
    'get_reaper_version',
    'get_resource_path',
    'get_user_inputs',
    'has_ext_state',
    'is_valid_id',
    'open_project',
    'perform_action',
    'prevent_ui_refresh',
    'print',
    'reaprint',
    'remove_reascript',
    'rgb_from_native',
    'rgb_to_native',
    'set_ext_state',
    'set_global_automation_mode',
    'show_console_message',
    'show_message_box',
    'slider_to_dB',
    'test_api',
    'undo_block',
    'update_arrange',
    'update_timeline',
    'view_prefs',
    'audio',
    'midi',
    'ui',
    'defer',
    'at_exit',
]



from typing import Tuple
import reapy_boost
from reapy_boost.core import ReapyObject
import reapy_boost.reascript_api as RPR


class Window(ReapyObject):

    def __init__(self, id: int) -> None:
        self.id = id

    @property
    def _args(self) -> Tuple[int]:
        return (self.id, )

    def refresh(self) -> None:
        """Refresh window."""
        RPR.DockWindowRefreshForHWND(self.id)  # type:ignore



from .marker import Marker, MarkerInfo
from .project import Project
from .region import Region, RegionInfo
from .time_selection import TimeSelection

__all__ = [
    "Marker",
    "MarkerInfo",
    "Project",
    "Region",
    "RegionInfo",
    "TimeSelection",
]



"""
Activate ``reapy_boost`` server.

Running this ReaScript from inside REAPER sets the ``reapy_boost`` server
that receives and executes API calls requests from outside. It will
automatically be run when importing ``reapy_boost`` from outside, if it is
enabled.
"""

import pathlib
import site
import sys

try:
    import reapy_boost
except ImportError:
    reapy_path = pathlib.Path(sys.path[0]).resolve().parent.parent
    sys.path.append(str(reapy_path))
    import reapy_boost

from reapy_boost.tools.network import Server


def run_main_loop() -> None:
    # Get new connections
    SERVER.accept()
    # Process API call requests
    requests = SERVER.get_requests()
    results = SERVER.process_requests(requests)
    SERVER.send_results(results)
    # Run main_loop again
    reapy_boost.defer(run_main_loop)


def get_new_reapy_server() -> Server:
    server_port = reapy_boost.config.REAPY_SERVER_PORT
    reapy_boost.set_ext_state("reapy_boost", "server_port", str(server_port))
    server = Server(server_port)
    return server


if __name__ == "__main__":
    SERVER = get_new_reapy_server()
    run_main_loop()
    reapy_boost.at_exit(
        reapy_boost.delete_ext_state, "reapy_boost", "server_port"
    )



"""
Enable ```reapy_boost`` distant API.

Running this ReaScript from inside REAPER allows to import ``reapy_boost``
from outside. It creates a persistent Web Interface inside REAPER and
adds the ReaScript ``reapy_boost.reascripts.activate_reapy_server`` to the
Actions list. Importing ``reapy_boost`` from outside REAPER will trigger
the latter **via** the Web Interface.
"""
import os
import traceback
from sys import exc_info
from sys import stdout

if __name__ == "__main__":
    try:
        import reapy_boost
    except Exception as e:
        traceback.print_tb(exc_info()[2], file=stdout)
        print(e)
    reapy_boost.config.enable_dist_api()
    from reapy_boost.core import _JS_generator
    api_filename = "_JS_API_generated.py"
    bin_dir = os.path.join(reapy_boost.get_resource_path(), "UserPlugins")
    _JS_generator.generate_js_api(bin_dir, api_filename)
    reapy_boost.generate_imgui()



import reapy_boost
from reapy_boost import reascript_api as RPR
from reapy_boost.core import ReapyObject
from reapy_boost.tools import depends_on_sws
import typing as ty


class Send(ReapyObject):
    """Track send.

    Attributes
    ----------
    index : int
        position on the track
    is_muted : bool
    is_phase_flipped : bool
    track_id : str
    type : str
        can be 'send', 'hardware' or 'receive'
    """

    _class_name = "Send"

    def __init__(
        self,
        track: ty.Optional['reapy_boost.Track'] = None,
        index: int = 0,
        track_id: ty.Optional[ty.Union[str, int]] = None,
        type: str = "send"
    ) -> None:
        if track_id is None:
            message = "One of `track` or `track_id` must be specified."
            assert track is not None, message
            track_id = track.id
        self.index = index
        self.track_id = track_id
        self.type = type

    def _get_int_type(self) -> int:
        types = {
            "hardware": 1,
            "send": 0,
            "receive": -1,
        }
        int_type = types[self.type]
        return int_type

    @property
    def _kwargs(self) -> ty.Dict[str, object]:
        return {
            "index": self.index,
            "track_id": self.track_id,
            "type": self.type
        }

    def delete(self) -> bool:  # type:ignore
        """
        Delete send.
        """
        RPR.RemoveTrackSend(  # type:ignore
            self.track_id, self._get_int_type(), self.index)

    @property
    def dest_track(self) -> 'reapy_boost.Track':
        """
        Destination track.

        :type: Track
        """
        pointer = self.get_info('P_DESTTRACK')
        track_id = reapy_boost.Track._get_id_from_pointer(pointer)
        return reapy_boost.Track(track_id)

    @reapy_boost.inside_reaper()
    def flip_phase(self) -> None:
        """
        Toggle whether phase is flipped.
        """
        self.is_phase_flipped = not self.is_phase_flipped

    def get_info(self, param_name: str) -> ty.Union[bool, int, str]:
        """Get raw info from GetTrackSendInfo_Value.

        Parameters
        ----------
        param_name : str
            B_MUTE : bool *
            B_PHASE : bool *, true to flip phase
            B_MONO : bool *
            D_VOL : double *, 1.0 = +0dB etc
            D_PAN : double *, -1..+1
            D_PANLAW : double *,1.0=+0.0db, 0.5=-6dB, -1.0 = projdef etc
            I_SENDMODE : int *, 0=post-fader, 1=pre-fx, 2=post-fx (deprecated),
                                3=post-fx
            I_AUTOMODE : int * : automation mode (-1=use track automode,
                                0=trim/off, 1=read, 2=touch, 3=write, 4=latch)
            I_SRCCHAN : int *, index,&1024=mono, -1 for none
            I_DSTCHAN : int *, index, &1024=mono, otherwise stereo pair,
                                hwout:&512=rearoute
            I_MIDIFLAGS : int *, low 5 bits=source channel 0=all, 1-16,
                                next 5 bits=dest channel, 0=orig,
                                1-16=chan
            P_DESTTRACK : read only, returns MediaTrack *,
                                destination track,
                                only applies for sends/recvs
            P_SRCTRACK : read only, returns MediaTrack *,
                                source track, only applies for sends/recvs
            P_ENV:<envchunkname : read only, returns TrackEnvelope *.
                                Call with :<VOLENV, :<PANENV, etc appended.


        Returns
        -------
        Union[bool, track id(str)]
        """
        value = RPR.GetTrackSendInfo_Value(  # type:ignore
            self.track_id, self._get_int_type(), self.index, param_name)
        return value

    @depends_on_sws
    def get_sws_info(self, param_name: str) -> float:
        """Raw value from BR_GetSetTrackSendInfo.

        Parameters
        ----------
        param_name : str
            B_MUTE : send mute state (1.0 if muted, otherwise 0.0)
            B_PHASE : send phase state (1.0 if phase is inverted, otherwise 0.0)
            B_MONO : send mono state (1.0 if send is set to mono, otherwise 0.0)
            D_VOL : send volume (1.0=+0dB etc...)
            D_PAN : send pan (-1.0=100%L, 0=center, 1.0=100%R)
            D_PANLAW : send pan law (1.0=+0.0db, 0.5=-6dB,
                        -1.0=project default etc...)
            I_SENDMODE : send mode (0=post-fader, 1=pre-fx, 2=post-fx(deprecated),
                        3=post-fx)
            I_SRCCHAN : audio source starting channel index or -1 if audio send
                        is disabled (&1024=mono...note that in that case, when
                        reading index, you should do (index XOR 1024) to get
                        starting channel index)
            I_DSTCHAN : audio destination starting channel index (&1024=mono
                        (and in case of hardware output &512=rearoute)...
                        note that in that case, when reading index, you should do
                        (index XOR (1024 OR 512)) to get starting channel index)
            I_MIDI_SRCCHAN : source MIDI channel, -1 if MIDI send is disabled
                            (0=all, 1-16)
            I_MIDI_DSTCHAN : destination MIDI channel, -1 if MIDI send is disabled
                            (0=original, 1-16)
            I_MIDI_SRCBUS : source MIDI bus, -1 if MIDI send is disabled
                            (0=all, otherwise bus index)
            I_MIDI_DSTBUS : receive MIDI bus, -1 if MIDI send is disabled
                            (0=all, otherwise bus index)
            I_MIDI_LINK_VOLPAN : link volume/pan controls to MIDI

        Returns
        -------
        float
        """
        value = RPR.BR_GetSetTrackSendInfo(  # type:ignore
            self.track_id, self._get_int_type(), self.index, param_name, False,
            0.0)
        return value

    @property
    def is_mono(self) -> bool:
        """
        Whether send is mono or stereo.

        :type: bool
        """
        is_mono = bool(self.get_info("B_MONO"))
        return is_mono

    @is_mono.setter
    def is_mono(self, mono: bool) -> None:
        self.set_info("B_MONO", mono)

    @property
    def is_muted(self) -> bool:
        """
        Whether send is muted.

        :type: bool
        """
        is_muted = bool(self.get_info("B_MUTE"))
        return is_muted

    @is_muted.setter
    def is_muted(self, is_muted: bool) -> None:
        """
        Mute or unmute send.

        Parameters
        ----------
        is_muted : bool
            Whether to mute or unmute send.
        """
        self.set_info("B_MUTE", is_muted)

    @property
    def is_phase_flipped(self) -> bool:
        """
        Whether send phase is flipped (i.e. signal multiplied by -1).

        :type: bool
        """
        is_phase_flipped = bool(self.get_info("B_PHASE"))
        return is_phase_flipped

    @is_phase_flipped.setter
    def is_phase_flipped(self, flipped: bool) -> None:
        self.set_info("B_PHASE", flipped)

    @property
    def _midi_flags_unpacked(
        self
    ) -> ty.Tuple[ty.Tuple[int, int], ty.Tuple[int, int]]:
        flags = int(self.get_info('I_MIDIFLAGS'))
        if flags == 0b1111111100000000011111:
            return ((-1, -1), (-1, -1))
        ch_flags = flags % 0b10000000000
        bus_flags = flags >> 14
        # bus
        src_bus = bus_flags % 0b100000
        dst_bus = bus_flags >> 8
        # channel
        src_ch = ch_flags % 0b100000
        dst_ch = ch_flags >> 5

        return ((src_bus, src_ch), (dst_bus, dst_ch))

    @_midi_flags_unpacked.setter
    def _midi_flags_unpacked(
        self, in_tuple: ty.Tuple[ty.Tuple[int, int], ty.Tuple[int, int]]
    ) -> None:
        src_bus, src_ch, dst_bus, dst_ch = (*in_tuple[0], *in_tuple[1])
        dst_ch <<= 5
        src_bus <<= 14
        dst_bus <<= 22
        flags = src_bus | src_ch | dst_bus | dst_ch
        if flags == -1:
            flags = 0b1111111100000000011111
        self.set_info('I_MIDIFLAGS', flags)

    @property
    def midi_source(self) -> ty.Tuple[int, int]:
        """
        Send MIDI properties on the send track.

        Returns
        -------
        Tuple[int bus, int channel]
        """
        return self._midi_flags_unpacked[0]

    @midi_source.setter
    def midi_source(self, source: ty.Tuple[int, int]) -> None:
        return self._midi_source_inside(source)

    @reapy_boost.inside_reaper()
    def _midi_source_inside(self, source: ty.Tuple[int, int]) -> None:
        dest = self._midi_flags_unpacked[1]
        self._midi_flags_unpacked = (source, dest)

    @property
    def midi_dest(self) -> ty.Tuple[int, int]:
        """
        Send MIDI properties on the receive track.

        Returns
        -------
        Tuple[int bus, int channel]
        """
        return self._midi_flags_unpacked[1]

    @midi_dest.setter
    def midi_dest(self, dest: ty.Tuple[int, int]) -> None:
        return self._midi_dest_inside(dest)

    @reapy_boost.inside_reaper()
    def _midi_dest_inside(self, dest: ty.Tuple[int, int]) -> None:
        source = self._midi_flags_unpacked[0]
        self._midi_flags_unpacked = (source, dest)

    def mute(self) -> None:
        """
        Mute send.
        """
        self.is_muted = True

    @property
    def pan(self) -> float:
        """
        Send pan (from -1=left to 1=right).

        :type: float
        """
        pan = ty.cast(float, self.get_info("D_PAN"))
        return pan

    @pan.setter
    def pan(self, pan: float) -> None:
        """
        Set send pan.

        Parameters
        ----------
        pan : float
            New pan between -1 (left) and 1 (right).
        """
        self.set_info("D_PAN", pan)

    def set_info(self, param_name: str, value: ty.Union[bool, float]) -> None:
        RPR.SetTrackSendInfo_Value(  # type:ignore
            self.track_id, self._get_int_type(), self.index, param_name, value)

    @depends_on_sws
    def set_sws_info(
        self, param_name: str, value: ty.Union[bool, float]
    ) -> None:
        RPR.BR_GetSetTrackSendInfo(  # type:ignore
            self.track_id, self._get_int_type(), self.index, param_name, True,
            value)

    @property
    def source_track(self) -> 'reapy_boost.Track':
        """
        Source track.

        :type: Track
        """
        pointer = ty.cast(str, self.get_info('P_SRCTRACK'))
        track_id = reapy_boost.Track._get_id_from_pointer(pointer)
        return reapy_boost.Track(track_id)

    def unmute(self) -> None:
        """
        Unmute send.
        """
        self.is_muted = False

    @property
    def volume(self) -> float:
        """
        Send volume.

        :type: float
        """
        volume = ty.cast(float, self.get_info("D_VOL"))
        return volume

    @volume.setter
    def volume(self, volume: float) -> None:
        self.set_info("D_VOL", volume)



from typing import List
import reapy_boost
from reapy_boost.tools import json

import os
import sys


@reapy_boost.inside_reaper()
def _get_api_names() -> List[str]:
    return __all__


if reapy_boost.is_inside_reaper():
    # Import functions without the useless starting "RPR_".
    import reaper_python as _RPR
    __all__ = [s[4:] for s in _RPR.__dict__ if s.startswith("RPR_")]
    for s in __all__:
        exec("{} = _RPR.__dict__['{}']".format(s, "RPR_" + s))

    from reapy_boost import additional_api as _A_API
    for s in _A_API.__dict__:
        exec("from reapy_boost.additional_api import {}".format(s))

    # Import SWS functions.
    try:
        sys.path.append(os.path.join(_RPR.RPR_GetResourcePath(), 'Scripts'))
        import sws_python as _SWS
        sws_functions = set(_SWS.__dict__) - set(_RPR.__dict__)
        __all__ += list(sws_functions)
        for s in sws_functions:
            exec("from sws_python import {}".format(s))
    except ImportError:  # SWS is not installed
        pass
else:
    if reapy_boost.dist_api_is_enabled():
        __all__ = _get_api_names()
        func_def = (
            "@reapy_boost.inside_reaper()\n"
            "def {name}(*args): return (name)(*args)"
        )
        exec("\n".join(func_def.format(name=name) for name in __all__))
        del func_def



from typing import List, Tuple, Union, cast
import reapy_boost
from reapy_boost import reascript_api as RPR
from reapy_boost.core import ReapyObject


class Item(ReapyObject):

    _class_name = "Item"

    def __init__(self, id: str):
        self.id = id

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Item) and self.id == other.id

    @property
    def _args(self) -> Tuple[str]:
        return self.id,

    @property
    def active_take(self) -> 'reapy_boost.Take':
        """
        Return the active take of the item.

        Returns
        -------
        take : Take
            Active take of the item.
        """
        take = reapy_boost.Take(RPR.GetActiveTake(self.id))  # type:ignore
        return take

    def add_take(self) -> 'reapy_boost.Take':
        """
        Create and return a new take in item.

        Returns
        -------
        take : Take
            New take in item.
        """
        take_id = RPR.AddTakeToMediaItem(self.id)  # type:ignore
        take = reapy_boost.Take(take_id)
        return take

    @reapy_boost.inside_reaper()
    def delete(self) -> None:
        """Delete item."""
        RPR.DeleteTrackMediaItem(self.track.id, self.id)  # type:ignore

    def get_info_value(self, param_name: str) -> float:
        value = RPR.GetMediaItemInfo_Value(self.id, param_name)  # type:ignore
        return value

    def get_take(self, index: int) -> 'reapy_boost.Take':
        """
        Return index-th take of item.

        Parameters
        ----------
        index : int
            Take index.

        Returns
        -------
        take : Take
            index-th take of media item.
        """
        take_id = RPR.GetItemTake(self.id, index)  # type:ignore
        take = reapy_boost.Take(take_id)
        return take

    @property
    def has_valid_id(self) -> bool:
        """
        Whether ReaScript ID is still valid.

        For instance, if item has been deleted, ID will not be valid
        anymore.

        :type: bool
        """
        return self._has_valid_id_inside()

    @reapy_boost.inside_reaper()
    def _has_valid_id_inside(self) -> bool:
        try:
            project_id = self.project.id
        except OSError:
            return False
        pointer, name = self._get_pointer_and_name()
        return bool(
            RPR.ValidatePtr2(  # type:ignore
                project_id, pointer, name
            )
        )

    @property
    def is_selected(self) -> bool:
        """
        Return whether item is selected.

        Returns
        -------
        is_selected : bool
            Whether item is selected.
        """
        is_selected = bool(RPR.IsMediaItemSelected(self.id))  # type:ignore
        return is_selected

    @is_selected.setter
    def is_selected(self, value: bool) -> None:
        self.set_info_value("B_UISEL", int(value))

    @property
    def length(self) -> float:
        """
        Return item length in seconds.

        Returns
        -------
        length : float
            Item length in seconds.
        """
        param_name = "D_LENGTH"
        length = self.get_info_value(param_name)
        return length

    @length.setter
    def length(self, length: float) -> None:
        """
        Set item length.

        Parameters
        ----------
        length : float
            New item length in seconds.
        """
        RPR.SetMediaItemLength(self.id, length, True)  # type:ignore

    @reapy_boost.inside_reaper()
    def make_only_selected_item(self) -> None:
        """
        Make track the only selected item in parent project.
        """
        self.project.select_all_items(False)
        self.is_selected = True

    @property
    def n_takes(self) -> int:
        """
        Return the number of takes of media item.

        Returns
        -------
        n_takes : int
            Number of takes of media item.
        """
        n_takes = cast(int, RPR.GetMediaItemNumTakes(self.id))  # type:ignore
        return n_takes

    @property
    def position(self) -> float:
        """
        Return item position in seconds.

        Returns
        -------
        position : float
            Item position in seconds.
        """
        position = self.get_info_value("D_POSITION")
        return position

    @position.setter
    def position(self, position: float) -> None:
        """
        Set media item position to `position`.

        Parameters
        ----------
        position : float
            New item position in seconds.
        """
        RPR.SetMediaItemPosition(self.id, position, False)  # type:ignore

    @property
    def project(self) -> 'reapy_boost.Project':
        """
        Item parent project.

        :type: reapy_boost.Project
        """
        return reapy_boost.Project(
            cast(
                str,
                RPR.GetItemProjectContext(  # type:ignore
                    self.id
                )
            )
        )

    def set_info_value(
        self, param_name: str, value: Union[bool, int, float]
    ) -> None:
        """
        Set raw item info value.

        Parameters
        ----------
        param_name : str
        value : float
        """
        RPR.SetMediaItemInfo_Value(self.id, param_name, value)  # type:ignore

    def split(self, position: float) -> Tuple['Item', 'Item']:
        """
        Split item and return left and right parts.

        Parameters
        ----------
        position : float
            Split position in seconds.

        Returns
        -------
        left, right : Item
            Left and right parts of the split.
        """
        right_id = RPR.SplitMediaItem(self.id, position)  # type:ignore
        left, right = self, Item(right_id)
        return left, right

    @property
    def takes(self) -> List['reapy_boost.Take']:
        """
        Return list of all takes of media item.

        Returns
        -------
        takes : list of Take
            List of all takes of media item.
        """
        return self._takes_inside()

    @reapy_boost.inside_reaper()
    def _takes_inside(self) -> List['reapy_boost.Take']:
        n_takes = RPR.GetMediaItemNumTakes(self.id)  # type:ignore
        take_ids = [
            RPR.GetMediaItemTake(  # type:ignore
                self.id, i
            ) for i in range(n_takes)
        ]
        takes = [reapy_boost.Take(take_id) for take_id in take_ids]
        return takes

    @property
    def track(self) -> 'reapy_boost.Track':
        """
        Parent track of item.

        Set it by passing a track, or a track index.

        :type: Track

        Examples
        --------
        >>> track0, track1 = project.tracks[0:2]
        >>> item = track0.items[0]
        >>> item.track == track0
        True
        >>> item.track = track1  # Move to track 1
        >>> item.track = 0  # Move to track 0
        """
        return self._track_inside()

    @reapy_boost.inside_reaper()
    def _track_inside(self) -> 'reapy_boost.Track':
        track_id = RPR.GetMediaItemTrack(self.id)  # type:ignore
        track = reapy_boost.Track(track_id)
        return track

    @track.setter
    def track(self, track: 'reapy_boost.Track') -> None:
        if isinstance(track, int):
            track = reapy_boost.Track(track, project=self.project)
        RPR.MoveMediaItemToTrack(self.id, track.id)  # type:ignore

    def update(self) -> None:
        """Update item in REAPER interface."""
        RPR.UpdateItemInProject(self.id)  # type:ignore



import struct
import typing as ty
from enum import IntFlag, IntEnum
import typing_extensions as te

import reapy_boost
import reapy_boost.reascript_api as RPR
from reapy_boost.core import ReapyObject, ReapyObjectList

T1 = ty.TypeVar('T1', bound='MIDIEvent')


class CCShapeFlag(IntFlag):
    square = 0
    linear = 16
    slow_start_end = 32
    fast_start = 16 | 32
    fast_end = 64
    beizer = 16 | 64

    @classmethod
    def from_shape(cls, shape: 'CCShape') -> 'CCShapeFlag':
        """
        Get flags from int enum.

        Parameters
        ----------
        shape : CCShape
            or just int from 0 to 5

        Returns
        -------
        CCShapeFlag
        """
        if shape == CCShape.square.value:
            return cls.square
        if shape == CCShape.linear.value:
            return cls.linear
        if shape == CCShape.slow_start_end.value:
            return cls.slow_start_end
        if shape == CCShape.fast_start.value:
            return cls.fast_start
        if shape == CCShape.fast_end.value:
            return cls.fast_end
        if shape == CCShape.beizer.value:
            return cls.beizer
        raise ValueError(f"Can't return flag for {shape}")


class CCShape(IntEnum):
    square = 0
    linear = 1
    slow_start_end = 2
    fast_start = 3
    fast_end = 4
    beizer = 5

    @classmethod
    def from_flag(cls, flag: CCShapeFlag) -> 'CCShape':
        """
        Get IntEnum from IntFlag.

        Parameters
        ----------
        flag : CCShapeFlag

        Returns
        -------
        CCShape
        """
        if flag == CCShapeFlag.square.value:
            return cls.square
        if flag == CCShapeFlag.linear.value:
            return cls.linear
        if flag == CCShapeFlag.slow_start_end.value:
            return cls.slow_start_end
        if flag == CCShapeFlag.fast_start.value:
            return cls.fast_start
        if flag == CCShapeFlag.fast_end.value:
            return cls.fast_end
        if flag == CCShapeFlag.beizer.value:
            return cls.beizer
        raise ValueError(f"Can't return shape for {flag}")


class MIDIEventDict(te.TypedDict):
    """
    Dict, represents trsferable midi event.

    Members
    -------
    ppq: float
    selected: bool
    muted: bool
    cc_shape: CCShapeFlag
    buf: ty.List[int]
    """

    ppq: float
    selected: bool
    muted: bool
    cc_shape: CCShapeFlag
    buf: ty.List[int]


class MIDIEventInfo(te.TypedDict):
    selected: bool
    muted: bool
    position: float
    raw_message: ty.List[int]
    ppq_position: float


class MIDIEvent(ReapyObject):
    """Abstract class for MIDI events."""

    _del_func_name = 'MIDI_DeleteEvt'
    index: int
    parent: 'reapy_boost.Take'

    def __init__(self, parent: 'reapy_boost.Take', index: int) -> None:
        """
        Create event.

        Parameters
        ----------
        parent : Take
            Take to which the event belongs.
        index : int
            Event index. It is specific to the event type, which means
            a Note and a CC can have the same index.
        """
        self.index = index
        self.parent = parent

    @property
    def _args(self) -> ty.Tuple['reapy_boost.Take', int]:
        return self.parent, self.index

    @property
    def as_dict(self) -> ty.Tuple[MIDIEventDict, ...]:
        """
        Representation of event in a dict.

        Notes
        -----
        ⋅ Can be used for inserting along with event buffer.
        ⋅ For notes and CC with beizer curve returns list of 2 items.

        :type: Tuple[MIDIEventDict]
        """
        infos = self.infos
        return (
            MIDIEventDict(
                ppq=infos['ppq_position'],
                selected=bool(infos['selected']),
                muted=bool(infos['muted']),
                cc_shape=CCShapeFlag(0),
                buf=infos['raw_message'],
            ),
        )

    def delete(self) -> None:
        """Delete event from the take."""
        f = getattr(RPR, self._del_func_name)
        f(self.parent.id, self.index)

    @property
    def infos(self) -> MIDIEventInfo:
        """
        Event info as dict.

        :type: MIDIEventInfo
            {'selected': bool,
            'muted': bool,
            'position': float,
            'message': List[int],
            'ppq_position': int}
        """
        take = self.parent
        max_eventbuf_length = 65000
        _, _, _, selected, muted, ppqpos, msg, _ = RPR.MIDI_GetEvt(
            take.id, self.index, 0, 0, 0.0, '', max_eventbuf_length
        )
        return MIDIEventInfo(
            selected=bool(selected),
            muted=bool(muted),
            position=take.ppq_to_time((ppqpos)),
            raw_message=[s for s in msg.encode('latin-1')],
            ppq_position=ppqpos
        )

    @reapy_boost.inside_reaper()
    def set(
        self,
        selected: ty.Optional[bool] = None,
        muted: ty.Optional[bool] = None,
        position: ty.Optional[float] = None,
        sort: ty.Optional[bool] = True,
        raw_message: ty.Optional[ty.List[int]] = None,
        *,
        time_unit: str = "seconds"
    ) -> None:
        """
        Set properties of event if needed.

        Note
        ----
        Optional arguments can be None, therefore will not be applied at all.

        Parameters
        ----------
        selected : bool, optional
            Whether to select new note (default=False).
        muted : bool, optional
            Whether to mute new note (default=False).
        position : float, optional
            position at take
        sort : bool, optional
            Whether to resort notes after creating new note
            (default=True). If False, then the new note will be
            ``take.notes[-1]``. Otherwise it will be at its place in
            the time-sorted list ``take.notes``. Set to False for
            improved efficiency when adding several notes, then call
            ``Take.sort_events`` at the end.
        raw_message : List[int], optional
            can be applied to every type of event, which leads to
            ignoring of the event-specific arguments.
        time_unit : str ("seconds as default")
            "beats"|"ppq"|"seconds" (default are seconds)
        """
        take = self.parent
        if position:
            position = take._resolve_midi_unit((position, ), time_unit)[0]
        if raw_message:
            raw_message = take._midi_to_bytestr(raw_message)
        RPR.MIDI_SetEvt(
            take.id, self.index, selected, muted, position, raw_message,
            len(raw_message), not sort
        )


class MIDIEventList(ReapyObjectList, ty.Generic[T1]):
    parent: 'reapy_boost.Take'

    def __init__(self, parent: 'reapy_boost.Take') -> None:
        """
        Create event list.

        Parameters
        ----------
        parent : Take
            Take to which the event list belongs.
        """
        self.parent = parent

    def __getitem__(self, key: int) -> T1:
        with reapy_boost.inside_reaper():
            if key >= len(self):
                raise IndexError
            return self._elements_class(self.parent, key)
        raise RuntimeError()

    def __len__(self) -> int:
        return ty.cast(int, getattr(self.parent, self._n_elements))

    def __iter__(self) -> ty.Iterator[T1]:
        for i in range(len(self)):
            yield self[i]

    @property
    def _args(self) -> ty.Tuple['reapy_boost.Take']:
        return self.parent,

    @property
    def _elements_class(self) -> ty.Type[T1]:
        return MIDIEvent

    @property
    def _n_elements(self) -> str:
        return 'n_midi_events'


class CCInfo(MIDIEventInfo):
    channel_message: int
    channel: int
    messages: ty.Tuple[int, int]


class CC(MIDIEvent):
    """MIDI CC event."""

    _del_func_name = 'MIDI_DeleteCC'

    @property
    def as_dict(self) -> ty.Tuple[MIDIEventDict, ...]:
        """
        Representation of event in a dict.

        Notes
        -----
        ⋅ Can be used for inserting along with event buffer.
        ⋅ For notes and CC with beizer curve returns list of 2 items.

        :type: Tuple[MIDIEventDict]
        """
        infos = self.infos
        shape = self.shape
        evt1 = MIDIEventDict(
            ppq=infos['ppq_position'],
            selected=bool(infos['selected']),
            muted=bool(infos['muted']),
            cc_shape=CCShapeFlag.from_shape(shape[0]),
            buf=infos['raw_message'],
        )
        if shape[0] == CCShape.beizer:
            evt2 = MIDIEventDict(
                ppq=infos['ppq_position'],
                selected=bool(infos['selected']),
                muted=bool(infos['muted']),
                cc_shape=CCShapeFlag.linear,
                buf=list(b'\xff\x0fCCBZ\x00' + struct.pack('<f', shape[1])),
            )
            return evt1, evt2
        return (evt1, )

    @property
    def channel(self) -> int:
        """
        CC channel between 0 and 15.

        :type: int

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        return self.infos['channel']

    @channel.setter
    def channel(self, channel: int) -> None:
        self.set(channel=channel)

    @property
    def channel_message(self) -> int:
        """
        CC channel message.

        :type: int

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        return self.infos['channel_message']

    @channel_message.setter
    def channel_message(self, channel_message: int) -> None:
        self.set(channel_message=channel_message)

    @reapy_boost.inside_reaper()
    @property
    def infos(self) -> CCInfo:
        """
        Return infos about CC.

        :type: CCInfo
            selected: bool
            muted: bool
            position: float
            ppq_position: float
            raw_message: List[int]
            channel_message: int
            channel: int
            messages: Tuple[int, int]
        """
        take = self.parent
        res = list(RPR.MIDI_GetCC(take.id, self.index, 0, 0, 0, 0, 0, 0,
                                  0))[3:]
        return CCInfo(
            selected=bool(res[0]),
            muted=bool(res[1]),
            position=take.ppq_to_time(res[2]),
            ppq_position=res[2],
            raw_message=list(int(b) for b in (res[3], *res[5:7])),
            channel_message=int(res[3]),
            channel=int(res[4]),
            messages=(int(res[5]), int(res[6])),
        )

    @property
    def messages(self) -> ty.Tuple[int, int]:
        """
        CC messages.

        :type: 2-list of int

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        return self.infos['messages']

    @messages.setter
    def messages(self, messages: ty.Tuple[int, int]) -> None:
        self.set(messages=messages)

    @property
    def muted(self) -> bool:
        """
        Whether CC is muted.

        :type: bool

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        return self.infos['muted']

    @muted.setter
    def muted(self, muted: bool) -> None:
        self.set(muted=muted)

    @property
    def position(self) -> float:
        """
        CC position in seconds.

        :type: float

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        return self.infos["position"]

    @position.setter
    def position(self, position: float) -> None:
        self.set(position=position)

    @property
    def selected(self) -> bool:
        """
        Whether CC is selected.

        :type: bool

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        return self.infos['selected']

    @selected.setter
    def selected(self, selected: bool) -> None:
        self.set(selected=selected)

    @reapy_boost.inside_reaper()
    def set(
        self,
        selected: ty.Optional[bool] = None,
        muted: ty.Optional[bool] = None,
        position: ty.Optional[float] = None,
        sort: ty.Optional[bool] = True,
        raw_message: ty.Optional[ty.List[int]] = None,
        channel_message: ty.Optional[int] = None,
        channel: ty.Optional[int] = None,
        messages: ty.Optional[ty.List[int]] = None,
        *,
        time_unit: str = "seconds"
    ) -> None:
        """
        Set properties of event if needed.

        Note
        ----
        Optional arguments can be None, therefore will not be applied at all.

        Parameters
        ----------
        selected : bool, optional
            Whether to select new note (default=False).
        muted : bool, optional
            Whether to mute new note (default=False).
        position : float, optional
            position at take
        sort : bool, optional
            Whether to resort notes after creating new note
            (default=True). If False, then the new note will be
            ``take.notes[-1]``. Otherwise it will be at its place in
            the time-sorted list ``take.notes``. Set to False for
            improved efficiency when adding several notes, then call
            ``Take.sort_events`` at the end.
        raw_message : List[int], optional
            can be applied to every type of event, which leads to
            ignoring of the event-specific arguments.
        channel_message : int, optional
            0xb0 for example
        channel : int, optional
        messages : Tuple[int, int], optional
            CC_num, value
        time_unit : str Optional
            "beats"|"ppq"|"seconds" (default are seconds)
        """
        take = self.parent
        if position:
            position = take._resolve_midi_unit((position, ), time_unit)[0]
        if sort is not None:
            sort = not sort
        if messages and not raw_message:
            msg2, msg3 = messages
        if raw_message:
            channel_message, msg2, msg3 = raw_message
            channel = None
        RPR.MIDI_SetCC(
            take.id, self.index, selected, muted, position, channel_message,
            channel, msg2, msg3, sort
        )

    @property
    def shape(self) -> ty.Tuple[CCShape, float]:
        """
        Shape type and beizer tension.

        :type: Tuple[CCShape, float]
            shape can be passed as enum instance or as int
        """
        _, _, _, shape_i, tension = RPR.MIDI_GetCCShape(
            self.parent.id, self.index, 1, 0.1
        )
        return CCShape(shape_i), tension

    @shape.setter
    def shape(self, shape: CCShape) -> None:
        shape_i = CCShape(shape[0]).value
        tension = shape[1]
        RPR.MIDI_SetCCShape(self.parent.id, self.index, shape_i, tension, None)


class CCList(MIDIEventList[CC]):

    _elements_class = CC
    _n_elements = "n_cc"


class NoteInfo(MIDIEventInfo):
    end: float
    ppq_end: float
    channel: int
    pitch: int
    velocity: int


class Note(MIDIEvent):
    """MIDI note."""

    _del_func_name = 'MIDI_DeleteNote'

    @property
    def as_dict(self) -> ty.Tuple[MIDIEventDict, MIDIEventDict]:
        """
        Representation of event in a dict.

        Notes
        -----
        ⋅ Can be used for inserting along with event buffer.
        ⋅ For notes and CC with beizer curve returns list of 2 items.

        :type: Tuple[MIDIEventDict, MIDIEventDict]
            note on and note off
        """
        infos = self.infos
        off = [infos['raw_message'][0] - 0x10, *infos['raw_message'][1:3]]
        return (
            MIDIEventDict(
                ppq=infos['ppq_position'],
                selected=bool(infos['selected']),
                muted=bool(infos['muted']),
                cc_shape=CCShapeFlag(0),
                buf=infos['raw_message'],
            ),
            MIDIEventDict(
                ppq=infos['ppq_end'],
                selected=bool(infos['selected']),
                muted=bool(infos['muted']),
                cc_shape=CCShapeFlag(0),
                buf=off,
            )
        )

    @property
    def channel(self) -> int:
        """
        Note channel between 0 and 15.

        :type: int

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        return self.infos['channel']

    @channel.setter
    def channel(self, channel: int) -> None:
        self.set(channel=channel)

    @property
    def end(self) -> float:
        """
        Note end in seconds.

        :type: float

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        return self.infos["end"]

    @end.setter
    def end(self, end: float) -> None:
        self.set(end=end)

    @property
    def infos(self) -> NoteInfo:
        """
        Return infos about note.

        :type: NoteInfo
            selected: bool
            muted: bool
            position: float
            ppq_position: float
            raw_message: [0x9+ch, note, vel]
            end: float
            ppq_end: float
            channel: int
            pitch: int
            velocity: int
        """
        take = self.parent
        res = list(RPR.MIDI_GetNote(take.id, self.index, 0, 0, 0, 0, 0, 0,
                                    0))[3:]
        ch, note, vel = int(res[4]), int(res[5]), int(res[6])
        return NoteInfo(
            selected=bool(res[0]),
            muted=bool(res[1]),
            position=take.ppq_to_time(res[2]),
            ppq_position=res[2],
            raw_message=[0x90 + ch, note, vel],
            end=take.ppq_to_time(res[3]),
            ppq_end=res[3],
            channel=ch,
            pitch=note,
            velocity=vel
        )

    @property
    def muted(self) -> bool:
        """
        Whether note is muted.

        :type: bool

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        return self.infos['muted']

    @muted.setter
    def muted(self, muted: bool) -> None:
        self.set(muted=muted)

    @property
    def pitch(self) -> int:
        """
        Note pitch between 0 and 127.

        :type: int

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        return self.infos['pitch']

    @pitch.setter
    def pitch(self, pitch: int) -> None:
        self.set(pitch=pitch)

    @property
    def selected(self) -> bool:
        """
        Whether note is selected.

        :type: bool

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        return self.infos['selected']

    @selected.setter
    def selected(self, selected: bool) -> None:
        self.set(selected=selected)

    @reapy_boost.inside_reaper()
    def set(
        self,
        selected: ty.Optional[bool] = None,
        muted: ty.Optional[bool] = None,
        position: ty.Optional[float] = None,
        sort: ty.Optional[bool] = True,
        raw_message: ty.Optional[ty.List[int]] = None,
        end: ty.Optional[float] = None,
        channel: ty.Optional[int] = None,
        pitch: ty.Optional[int] = None,
        velocity: ty.Optional[int] = None,
        *,
        time_unit: str = "seconds"
    ) -> None:
        """
        Set properties of event if needed.

        Note
        ----
        Optional arguments can be None, thereforce will not be applied at all.

        Parameters
        ----------
        selected : bool, optional
            Whether to select new note (default=False).
        muted : bool, optional
            Whether to mute new note (default=False).
        position : float, optional
            position at take
        sort : bool, optional
            Whether to resort notes after creating new note
            (default=True). If False, then the new note will be
            ``take.notes[-1]``. Otherwise it will be at its place in
            the time-sorted list ``take.notes``. Set to False for
            improved efficiency when adding several notes, then call
            ``Take.sort_events`` at the end.
        raw_message : List[int], optional
            can be applied to every type of event, which leads to
            ignoring of the event-specific arguments.
        end : float
        channel : int
        pitch : int
        velocity : int
        time_unit : str ("seconds as default")
            "beats"|"ppq"|"seconds" (default are seconds)
        """
        take = self.parent
        if position:
            position = take._resolve_midi_unit((position, ), time_unit)[0]
        if end:
            end = take._resolve_midi_unit((end, ), time_unit)[0]
        if raw_message:
            rm = raw_message
            channel, pitch, velocity = rm[0] % 0x90, rm[1], rm[2]
        if sort is not None:
            sort = not sort
        RPR.MIDI_SetNote(
            take.id, self.index, selected, muted, position, end, channel,
            pitch, velocity, sort
        )

    @property
    def start(self) -> float:
        """
        Note start in seconds.

        :type: float

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        return self.infos["position"]

    @start.setter
    def start(self, start: float) -> None:
        self.set(position=start)

    @property
    def velocity(self) -> int:
        """
        Note velocity between 0 and 127.

        :type: int

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        return self.infos['velocity']

    @velocity.setter
    def velocity(self, velocity: int) -> None:
        self.set(velocity=velocity)


class NoteList(MIDIEventList[Note]):

    _elements_class = Note
    _n_elements = "n_notes"


class TextSysexInfo(MIDIEventInfo):
    type_: int


class TextSysex(MIDIEvent):
    """Abstract class for Text or Sysex events."""

    _del_func_name = 'MIDI_DeleteTextSysexEvt'

    @property
    def infos(self) -> TextSysexInfo:
        """
        Event info as dict.

        :type: TextSysexInfo
            {'selected': bool,
            'muted': bool,
            'position': float,
            'message': List[int],
            'ppq_position': float,
            'type_': int
            }
        """
        take = self.parent
        max_eventbuf_length = 65000
        _, _, _, sel, muted, ppqpos, type_, msg, _ = RPR.MIDI_GetTextSysexEvt(
            take.id, self.index, 0, 0, 0.0, 0, '', max_eventbuf_length
        )
        return TextSysexInfo(
            selected=bool(sel),
            muted=bool(muted),
            position=take.ppq_to_time((ppqpos)),
            raw_message=[s for s in msg.encode('latin-1')],
            ppq_position=ppqpos,
            type_=int(type_)
        )

    @reapy_boost.inside_reaper()
    def set(
        self,
        selected: ty.Optional[bool] = None,
        muted: ty.Optional[bool] = None,
        position: ty.Optional[float] = None,
        sort: ty.Optional[bool] = True,
        raw_message: ty.Optional[ty.List[int]] = None,
        type_: ty.Optional[int] = None,
        *,
        time_unit: str = "seconds"
    ) -> None:
        """
        Set properties of event if needed.

        Notes
        -----
        ⋅ Optional arguments can be None, therefore will not be applied at all
        ⋅ type_ is NOT optional when new raw_message is assigned

        Parameters
        ----------
        selected : bool, optional
            Whether to select new note (default=False).
        muted : bool, optional
            Whether to mute new note (default=False).
        position : float, optional
            position at take
        sort : bool, optional
            Whether to resort notes after creating new note
            (default=True). If False, then the new note will be
            ``take.notes[-1]``. Otherwise it will be at its place in
            the time-sorted list ``take.notes``. Set to False for
            improved efficiency when adding several notes, then call
            ``Take.sort_events`` at the end.
        raw_message : List[int], optional
            can be applied to every type of event, which leads to
            ignoring of the event-specific arguments.
        type_ : int, optional
            -1:sysex (msg should not include bounding F0..F7),
            1-14:MIDI text event types,
            15=REAPER notation event.
        time_unit : str ("seconds as default")
            "beats"|"ppq"|"seconds" (default are seconds)
        """
        take = self.parent
        print('called')
        if position:
            position = take._resolve_midi_unit((position, ), time_unit)[0]
        if raw_message:
            raw_message = take._midi_to_bytestr(raw_message)
            print(raw_message, len(raw_message))
        RPR.MIDI_SetTextSysexEvt(
            take.id, self.index, selected, muted, position, type_, raw_message,
            len(raw_message), not sort
        )

    @property
    def type_(self) -> int:
        """
        Meta event type

        :type: int
            -1:sysex (msg should not include bounding F0..F7),
            1-14:MIDI text event types,
            15=REAPER notation event.
        """
        return self.infos['type_']

    @type_.setter
    def type_(self, type_: int) -> None:
        self.set(type_=type_)


class TextSysexList(MIDIEventList[TextSysex]):

    _elements_class = TextSysex
    _n_elements = "n_text_sysex"



import functools

from reapy_boost.errors import ExtensionNotFoundError

import reapy_boost

from pathlib import Path
from warnings import warn
import typing as ty
import re

FuncType = ty.Callable[..., ty.Any]
F = ty.TypeVar('F', bound=FuncType)


def depends_on_extension(extension: str, url: str) -> ty.Callable[[F], F]:
    """Return a decorator to indicate dependency to an extension.

    If the extension is not available, an `ExtensionNotFoundError`
    will be raised when calling the decorated function.

    Parameters
    ----------
    extension : str
        Extension name.
    url : str
        URL of the download page or installation instructions of
        the extension.
    """
    message = "module 'reapy_boost.reascript_api' has no attribute"

    def decorator(f: F) -> F:
        """Indicate dependency of a function to an extension.

        If the extension is not available, an `ExtensionNotFoundError`
        will be raised when calling the decorated function.
        """

        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except AttributeError as exc:
                if exc.args[0].startswith(message):
                    raise ExtensionNotFoundError(extension, url)
                else:
                    raise exc

        return wrapped

    return decorator


depends_on_sws = depends_on_extension('SWS', 'www.sws-extension.org')

_preambula = """
import reapy_boost
if reapy_boost.is_inside_reaper():
    from reaper_python import *\n"""
_decorator = """@reapy_boost.inside_reaper()\n"""


def generate_imgui() -> None:
    """Generate API for dear ImGui.

    Note
    ----
    Extension ReaImGui (https://github.com/cfillion/reaimgui)
    has to be installed.
    """
    api_path = Path(
        reapy_boost.get_resource_path()
    ).joinpath("Scripts/ReaTeam Extensions/API/imgui_python.py")
    if not api_path.exists():
        warn(
            "ImGUI API path does not exists." +
            " Install the extension with ReaPack"
        )
        return
    new_path = Path(__file__).parent.parent.joinpath('ImGui.py')
    assert new_path.exists()
    with open(new_path, 'w') as new:
        new.write(_preambula)
        with open(api_path, 'r') as old:
            for line in old.readlines():
                if line.startswith("from reaper_python import *"):
                    continue
                line = re.sub(r'def ImGui_', _decorator + 'def ', line)
                if not re.search(r"rpr_getfp\('ImGui_", line):
                    line = re.sub(r'ImGui_', '', line)
                # newlines.append(line)
                new.write(line)



"""Helper module to find python shared library.

This is a modified version of the script located at
https://gist.github.com/tkf/d980eee120611604c0b9b5fef5b8dae6.

Below are the original copyright and license:

    Copyright 2018, Takafumi Arakaki

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import ctypes.util
import ctypes
import os
import sys
import sysconfig
from typing import Generator, List, Optional, Union, cast

__all__ = ["get_python_shared_library"]


class _Dl_info(ctypes.Structure):

    _fields_ = [
        ("dli_fname", ctypes.c_char_p),
        ("dli_fbase", ctypes.c_void_p),
        ("dli_sname", ctypes.c_char_p),
        ("dli_saddr", ctypes.c_void_p),
    ]


def get_candidate_names() -> Generator[str, None, None]:
    """Yield candidate file names for shared lib."""
    suffix = get_sharedlib_suffix()
    LDLIBRARY = sysconfig.get_config_var("LDLIBRARY")
    if LDLIBRARY:
        yield LDLIBRARY

    LIBRARY = sysconfig.get_config_var("LIBRARY")
    if LIBRARY:
        yield os.path.splitext(LIBRARY)[0] + suffix

    dlprefix = "" if is_windows() else "lib"
    sysdata = dict(
        v=sys.version_info,
        # VERSION is X.Y in Linux/macOS and XY in Windows:
        VERSION=(
            sysconfig.get_config_var("VERSION") or
            "{v.major}.{v.minor}".format(v=sys.version_info)
        ),
        ABIFLAGS=(
            sysconfig.get_config_var("ABIFLAGS") or
            sysconfig.get_config_var("abiflags") or ""
        ),
    )

    for stem in (
        "python{VERSION}{ABIFLAGS}".format(**sysdata),
        "python{VERSION}".format(**sysdata),
        "python{v.major}".format(**sysdata), "python"
    ):
        yield dlprefix + stem + suffix


def get_candidate_paths() -> Generator[str, None, None]:
    """Yield candidate paths to shared lib."""
    if pyth := get_linked_libpython():
        yield pyth

    # List candidates for directories in which libpython may exist
    config_vars = "LIBPL", "srcdir", "LIBDIR"
    lib_dirs: List[Union[str, None]] = list(
        map(sysconfig.get_config_var, config_vars)
    )

    if is_windows():
        lib_dirs.append(os.path.join(os.path.dirname(sys.executable)))
    else:
        lib_dirs.append(
            os.path.join(
                os.path.dirname(os.path.dirname(sys.executable)), "lib"
            )
        )

    # For macOS:
    lib_dirs.append(sysconfig.get_config_var("PYTHONFRAMEWORKPREFIX"))

    lib_dirs.append(sys.exec_prefix)
    lib_dirs.append(os.path.join(sys.exec_prefix, "lib"))

    lib_basenames = list(get_candidate_names())

    for directory in filter(bool, lib_dirs):
        for basename in lib_basenames:
            yield os.path.join(cast(str, directory), basename)

    # In macOS and Windows, ctypes.util.find_library returns a full path:
    for basename in lib_basenames:
        lib = ctypes.util.find_library(get_library_name(basename))
        if not lib:
            raise RuntimeError('cannot find Python library')
        yield lib


def get_library_name(name: str) -> str:
    """Convert file name to library name (no "lib" and ".so" etc.)."""
    suffix = get_sharedlib_suffix()
    if not is_windows() and name.startswith("lib"):
        name = name[len("lib"):]
    if suffix and name.endswith(suffix):
        name = name[:-len(suffix)]
    return name


def get_linked_libpython() -> Optional[str]:
    """Return linked libpython using dladdr (in *nix).

    Return ``None`` if libpython is statically linked.
    """
    if is_windows():
        return
    libdl = ctypes.CDLL(ctypes.util.find_library("dl"))
    libdl.dladdr.argtypes = [ctypes.c_void_p, ctypes.POINTER(_Dl_info)]
    libdl.dladdr.restype = ctypes.c_int

    dlinfo = _Dl_info()
    retcode = libdl.dladdr(
        ctypes.cast(ctypes.pythonapi.Py_GetVersion, ctypes.c_void_p),
        ctypes.pointer(dlinfo)
    )
    if retcode == 0:  # means error
        return
    path = os.path.realpath(dlinfo.dli_fname.decode())
    if path == os.path.realpath(sys.executable):
        return
    return path


def get_python_shared_library() -> str:
    """Return path to Python shared library (.dll, .dylib or .so).

    Returns
    -------
    path : str
        Path to the (hopefully) correct libpython.

    Raises
    ------
    FileNotFoundError
        When shared library could not be found.
    """
    for path in filter(is_valid, get_candidate_paths()):
        return path
    raise FileNotFoundError(
        "Could not find Python shared library. Please report this bug at " +
        "https://github.com/RomeoDespres/reapy_boost/issues/new so that we can "
        + "support more cases."
    )


def get_sharedlib_suffix() -> str:
    """Return shared library suffix (.dll, .dylib or .so)."""
    suffix = sysconfig.get_config_var("SHLIB_SUFFIX")
    if suffix is None:
        if is_windows():
            suffix = ".dll"
        else:
            suffix = ".so"
    if is_apple():
        # sysconfig.get_config_var("SHLIB_SUFFIX") can be ".so" in macOS.
        # Let's not use the value from sysconfig.
        suffix = ".dylib"
    return suffix


def is_apple() -> bool:
    """Return whether OS is MacOS or OSX."""
    return sys.platform == "darwin"


def is_valid(path: str) -> bool:
    """Return whether path is a valid library path."""
    return (
        bool(path) and os.path.isabs(path) and os.path.exists(path) and
        (not is_apple() or path.endswith(".dylib"))
    )


def is_windows() -> bool:
    """Return whether OS is Windows."""
    return os.name == "nt"



from setuptools import setup, find_packages
from os import path, walk, sep
from typing import Dict, List

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md")) as f:
    long_description = f.read()


def find_stubs(package: str) -> Dict[str, List[str]]:
    stubs = []
    for root, dirs, files in walk(package):
        for file in files:
            r_path = path.join(root, file).replace(package + sep, '', 1)
            stubs.append(r_path)
    stubs.append('py.typed')
    return {package: stubs}


with open(path.join(here, 'reapy_boost', '__init__.py')) as f:
    for line in f.readlines():
        line = line.strip()
        if line.startswith('__version__'):
            version = line.split('"')[1]

setup(
    name="reapy-boost",
    version=version,
    description="A pythonic wrapper for REAPER's ReaScript Python API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Roméo Després, forked by Levitanus",
    author_email="pianoist@ya.ru",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3"
    ],
    keywords="REAPER DAW ReaScript API wrapper music audio",
    packages=find_packages(exclude=["docs"]),
    package_data=find_stubs('reapy_boost'),
    install_requires=[
        'psutil',
    ],
    python_requires=">=3.7"
)



from .reapy_object import ReapyObject, ReapyObjectList

from .audio_accessor import AudioAccessor
from .envelope import Envelope, EnvelopeList, EnvelopePoint
from .fx import FX, FXList, FXParam, FXParamsList
from .item import (Item, MIDIEvent, MIDIEventList, CC, CCList, Note, NoteList,
                   TextSysex, TextSysexInfo, TextSysexList, CCShapeFlag,
                   CCShape, MIDIEventDict, MIDIEventInfo, CCInfo, NoteInfo,
                   Source, Take)
from .map import map
from .track import AutomationItem, Send, Track, TrackList
from .project import (Marker, MarkerInfo, Project, Region, RegionInfo,
                      TimeSelection)
from .window import MIDIEditor, ToolTip, Window
from . import JS_API as JS

__all__ = [
    # core.reapy_object
    "ReapyObject",
    "ReapyObjectList",
    # core.audio_accessor
    "AudioAccessor",
    # core.envelope
    "Envelope",
    "EnvelopeList",
    "EnvelopePoint",
    # core.fx
    "FX",
    "FXList",
    "FXParam",
    "FXParamsList",
    # core.item
    "Item",
    "MIDIEvent",
    "MIDIEventList",
    "CC",
    "CCList",
    "Note",
    "NoteList",
    "TextSysex",
    "TextSysexInfo",
    "TextSysexList",
    "CCShapeFlag",
    "CCShape",
    "MIDIEventDict",
    "MIDIEventInfo",
    "CCInfo",
    'NoteInfo',
    "Source",
    "Take",
    # core.map
    "map",
    # core.project
    "Marker",
    "MarkerInfo",
    "Project",
    "Region",
    "RegionInfo",
    "TimeSelection",
    # core.track
    "AutomationItem",
    "Send",
    "Track",
    "TrackList",
    # core.window
    "MIDIEditor",
    "ToolTip",
    "Window",
    # JS_API
    "JS",
]



"""User interface-related functions."""

import reapy_boost
import reapy_boost.reascript_api as RPR


def get_color_theme() -> str:
    """
    Return path to last color theme file.

    Returns
    -------
    color_theme : str
        Path to last color theme file.
    """
    return RPR.GetLastColorThemeFile()  # type:ignore


def get_leftmost_track_in_mixer() -> 'reapy_boost.Track':
    """Return leftmost track in mixer."""
    return reapy_boost.Track(RPR.GetMixerScroll())  # type:ignore


def set_color_theme(path: str) -> None:
    """
    Set REAPER color theme.

    Parameters
    ----------
    path : str
        Path to color theme file.
    """
    RPR.OpenColorThemeFile(path)  # type:ignore


def set_leftmost_track_in_mixer(track: 'reapy_boost.Track') -> None:
    """Set leftmost track in mixer.

    The actual leftmost track in mixer might be different after
    calling if there is not enough tracks to the right of the
    specified track.

    Parameters
    ----------
    track : Track
        Track to set as leftmost track in mixer.
    """
    RPR.SetMixerScroll(track.id)  # type:ignore



"""Encode and decode ``reapy_boost`` objects as JSON."""

import importlib
import inspect
import json
import operator
import sys
import typing as ty
import types
if ty.TYPE_CHECKING:
    import reapy_boost

T1 = ty.TypeVar('T1')


class ClassCache(ty.Dict[str, T1]):

    _core: ty.Optional[types.ModuleType] = None

    def __missing__(self, key: str) -> T1:
        if self._core is None:
            # The import is here because otherwise there is an import loop
            # and to perform import just once.
            self._core = importlib.import_module("reapy_boost.core")
        self[key] = getattr(self._core, key)
        return self[key]


_CLASS_CACHE: ClassCache[ty.Type['reapy_boost.ReapyObject']] = ClassCache()


class ReapyEncoder(json.JSONEncoder):

    def default(self, x: ty.Any) -> ty.Any:
        if hasattr(x, '_to_dict'):
            return x._to_dict()
        elif inspect.ismethod(x):
            return {"__self__": x.__self__, "__method_name__": x.__name__}
        elif callable(x):
            return {
                "__callable__": True,
                "module_name": x.__module__,
                "name": x.__qualname__
            }
        elif isinstance(x, slice):
            return {"__slice__": True, "args": (x.start, x.stop, x.step)}
        return json.JSONEncoder.default(self, x)


def loads(s: ty.Union[ty.Text, bytes]) -> ty.Any:
    return json.loads(s, object_hook=object_hook)


def dumps(x: ty.Any) -> str:
    return json.dumps(x, cls=ReapyEncoder)


def object_hook(x: ty.Any) -> ty.Any:
    if "__reapy__" in x:
        reapy_class = _CLASS_CACHE[x["class"]]
        return reapy_class(*x["args"], **x["kwargs"])
    elif "__method_name__" in x:
        return getattr(x["__self__"], x["__method_name__"])
    elif "__callable__" in x:
        module_name, name = x["module_name"], x["name"]
        try:
            module = sys.modules[module_name]
        except KeyError:
            module = importlib.import_module(module_name)
        return operator.attrgetter(name)(sys.modules[module_name])
    elif "__slice__" in x:
        return slice(*x["args"])
    else:
        return x



import reapy_boost
from reapy_boost.core.item.take import Take
import reapy_boost.reascript_api as RPR
from .window import Window


class MIDIEditor(Window):

    def _get_int_setting(self, setting: str) -> int:
        return RPR.MIDIEditor_GetSetting_int(self.id, setting)  # type:ignore

    def _get_str_setting(self, setting: str) -> str:
        return RPR.MIDIEditor_GetSetting_str(  # type:ignore
            self.id, setting, "", 2048)[3]

    @property
    def last_clicked_cc_lane(self) -> int:
        """
        Last clicked CC lane.

        :type: int
        """
        return self._get_int_setting("last_clicked_cc_lane")

    @property
    def last_clicked_cc_lane_name(self) -> str:
        """
        Last clicked CC lane name ("velocity", "pitch", etc.).

        :type: str
        """
        return self._get_str_setting("last_clicked_cc_lane")

    @property
    def active_note_row(self) -> int:
        """
        Active note row (between 0 and 127).

        :type: int
        """
        return self._get_int_setting("active_note_row")

    @property
    def default_channel(self) -> int:
        """
        Default note channel (between 0 and 15).

        :type: int
        """
        return self._get_int_setting("default_note_chan")

    @property
    def default_length(self) -> int:
        """
        Default note length in MIDI ticks.

        :type: int
        """
        return self._get_int_setting("default_note_len")

    @property
    def default_velocity(self) -> int:
        """
        Default note velocity (between 0 and 127).

        :type: int
        """
        return self._get_int_setting("default_note_vel")

    @property
    def is_scale_enabled(self) -> bool:
        """
        Whether scale is enabled in editor.

        :type: bool
        """
        return bool(self._get_int_setting("scale_enabled"))

    @property
    def is_snap_enabled(self) -> bool:
        """
        Whether snap is enabled in editor.

        :type: bool
        """
        return bool(self._get_int_setting("snap_enabled"))

    @property
    def mode(self) -> str:
        """
        Mode of MIDI editor.

        :type: {0: "piano roll", 1: "event list"}
        """
        modes = {0: "piano roll", 1: "event list"}
        return modes[RPR.MIDIEditor_GetMode(self.id)]  # type:ignore

    def perform_action(self, action_id: int) -> None:
        """
        Perform action (from MIDI Editor section).

        Parameters
        ----------
        action_id : int
            Action ID.
        """
        RPR.MIDIEditor_OnCommand(self.id, action_id)  # type:ignore

    @property
    def scale_type(self) -> str:
        """
        Scale type ID.

        :type: str
        """
        return self._get_str_setting("scale")

    @property
    def scale_root(self) -> int:
        """
        Scale root (between 0 and 12, 0=C).

        :type: int
        """
        return self._get_int_setting("scale_root")

    @property
    def take(self) -> Take:
        """
        Take currently edited.

        :type: Take
        """
        return reapy_boost.Take(RPR.MIDIEditor_GetTake(self.id))  # type:ignore



from typing_extensions import TypedDict
from typing import Dict, List, Optional, Union, cast
import reapy_boost
from reapy_boost import reascript_api as RPR
from reapy_boost.core import ReapyObject


class RegionInfo(TypedDict):
    index: int
    enum_index: int
    project_id: str
    name: str
    start: float
    end: float
    rendered_tracks: List['reapy_boost.Track']


class Region(ReapyObject):

    _class_name = "Region"
    project_id: str
    index: int
    enum_index: int

    def __init__(
        self,
        parent_project: Optional['reapy_boost.Project'] = None,
        index: Optional[int] = None,
        parent_project_id: Optional[str] = None,
        enum_index: Optional[int] = None
    ) -> None:
        if parent_project_id is None:
            message = (
                "One of `parent_project` or `parent_project_id` must be "
                "specified."
            )
            assert parent_project is not None, message
            parent_project_id = parent_project.id
        self.project = reapy_boost.Project(parent_project_id)
        self.project_id = parent_project_id
        if index is None:
            if enum_index is None:
                index = self.project.n_regions
            else:
                index = cast(
                    int,
                    RPR.EnumProjectMarkers2(  # type:ignore
                        self.project_id, enum_index, 0, 0, 0, 0, 0
                    )[7]
                )
        self.index = index
        if enum_index is None:
            # self.enum_index = None
            enum_index = self._get_enum_index()
        self.enum_index = enum_index

    @reapy_boost.inside_reaper()
    def _get_enum_index(self) -> int:
        """
        Return region index as needed by RPR.EnumProjectMarkers2.

        Raises
        ------
        reapy_boost.errors.UndefinedRegionError
        """
        for region in self.project.regions:
            if region.index == self.index:
                return region.enum_index
        raise reapy_boost.errors.UndefinedRegionError(self.index)

    @property
    def _kwargs(self) -> Dict[str, Union[str, int]]:
        return {
            "index": self.index,
            "parent_project_id": self.project_id,
            "enum_index": self.enum_index
        }

    def add_rendered_track(self, track: 'reapy_boost.Track') -> None:
        """
        Add track to region render matrix for this region.

        Parameters
        ----------
        track : Track
            Track to add.

        See also
        --------
        Region.add_rendered_tracks
            Efficiently add several tracks to region render matrix.
        Region.remove_rendered_track
        Region.remove_rendered_tracks
        """
        RPR.SetRegionRenderMatrix(  # type:ignore
            self.project_id, self.index, track.id, 1)

    @reapy_boost.inside_reaper()
    def add_rendered_tracks(self, tracks: List['reapy_boost.Track']) -> None:
        """
        Efficiently add  several tracks to region render matrix.

        Parameters
        ----------
        tracks : list of Track
            Tracks to add.

        See also
        --------
        Region.remove_rendered_tracks
        """
        for track in tracks:
            self.add_rendered_track(track)

    @property
    def end(self) -> float:
        """
        Region end.

        :type: float
            Region end in seconds.
        """
        return self._end_inside()

    @reapy_boost.inside_reaper()
    def _end_inside(self) -> float:
        index = self._get_enum_index()
        args = self.project_id, index, 0, 0, 0, 0, 0
        return cast(float, RPR.EnumProjectMarkers2(*args)[5])  # type:ignore

    @end.setter
    def end(self, end: float) -> None:
        """
        Set region end.

        Parameters
        ----------
        end : float
            region end in seconds.
        """
        args = self.project_id, self.index, True, self.start, end, self.name
        RPR.SetProjectMarker2(*args)  # type:ignore

    def delete(self) -> None:
        """
        Delete region.
        """
        RPR.DeleteProjectMarker(  # type:ignore
            self.project_id, self.index, True)

    @property
    def name(self) -> str:
        """
        Region name.

        :type: str
        """
        return self._name_inside()

    @reapy_boost.inside_reaper()
    def _name_inside(self) -> str:
        # index = self._get_enum_index()
        fs = RPR.SNM_CreateFastString('0' * 1024)  # type:ignore
        args = self.project_id, self.index, True, fs
        RPR.SNM_GetProjectMarkerName(*args)  # type:ignore
        result = cast(str, RPR.SNM_GetFastString(fs))  # type:ignore
        RPR.SNM_DeleteFastString(fs)  # type:ignore
        return result

    @name.setter
    def name(self, name: str) -> None:
        """
        Set region name.

        Parameters
        ----------
        name : str
        """
        args = self.project_id, self.index, True, self.start, self.end, name
        RPR.SetProjectMarker2(*args)  # type:ignore

    def remove_rendered_track(self, track: 'reapy_boost.Track') -> None:
        """
        Remove track from region render matrix for this region.

        Parameters
        ----------
        track : Track
            Track to remove.

        See also
        --------
        Region.add_rendered_tracks
        Region.remove_rendered_track
        Region.remove_rendered_tracks
            Efficiently remove several tracks from render matrix.
        """
        RPR.SetRegionRenderMatrix(  # type:ignore
            self.project_id, self.index, track.id, -1)

    @reapy_boost.inside_reaper()
    def remove_rendered_tracks(
        self, tracks: List['reapy_boost.Track']
    ) -> None:
        """
        Efficiently remove  several tracks from region render matrix.

        Parameters
        ----------
        tracks : list of Track
            Tracks to remove.

        See also
        --------
        Region.add_rendered_tracks
        """
        for track in tracks:
            self.remove_rendered_track(track)

    @property
    def rendered_tracks(self) -> List['reapy_boost.Track']:
        """
        List of tracks for this region in region render matrix.

        :type: list of Track
        """
        return self._rendered_tracks_inside()

    @reapy_boost.inside_reaper()
    def _rendered_tracks_inside(self) -> List['reapy_boost.Track']:
        i = 0
        tracks = []
        while i == 0 or tracks[-1]._is_defined:
            track_id = RPR.EnumRegionRenderMatrix(  # type:ignore
                self.project_id, self.index, i
            )
            tracks.append(reapy_boost.Track(track_id))
            i += 1
        return tracks[:-1]

    @property
    def start(self) -> float:
        """
        Region start.

        :type: float
        """
        return self._start_inside()

    @reapy_boost.inside_reaper()
    def _start_inside(self) -> float:
        args = self.project_id, self._get_enum_index(), 0, 0, 0, 0, 0
        return RPR.EnumProjectMarkers2(*args)[4]  # type:ignore

    @start.setter
    def start(self, start: float) -> None:
        """
        Set region start.

        Parameters
        ----------
        start : float
            region start in seconds.
        """
        RPR.SetProjectMarker2(  # type:ignore
            self.project_id, self.index, 1, start, self.end, self.name
        )

    @property
    def infos(self) -> RegionInfo:
        """Get all Region infos in one call.

        Returns
        -------
        RegionInfo
            index: int
            enum_index: int
            project_id: str
            name: str
            start: float
            end: float
            rendered_tracks: List[reapy_boost.Track]
        """
        return self._infos_inside()

    @reapy_boost.inside_reaper()
    def _infos_inside(self) -> RegionInfo:
        enum_index = self._get_enum_index()
        args = self.project_id, enum_index, 0, 0, 0, 0, 0
        (_, _, _, _, start, end, _,
         index) = RPR.EnumProjectMarkers2(  # type:ignore
             *args
         )
        out: RegionInfo = {
            'index': index,
            'enum_index': enum_index,
            'project_id': self.project_id,
            'start': start,
            'end': end,
            'name': self.name,
            'rendered_tracks': self.rendered_tracks
        }
        return out



from reapy_boost.errors import DisabledDistAPIError, UndefinedExtStateError
from reapy_boost.tools import json
from urllib import request
from urllib.error import URLError

import reapy_boost
import typing as ty


class WebInterface:
    _url: str
    ext_state: ExtState

    def __init__(self, port: int, host: str = 'localhost') -> None:
        ...

    def activate_reapy_server(self) -> None:
        ...

    def get_reapy_server_port(self) -> int:
        ...

    def perform_action(self, action_id: ty.Union[int, str, bytes]) -> None:
        ...


class ExtState:
    _url: str

    def __init__(self, web_interface: WebInterface) -> None:
        ...

    def __getitem__(self, key: str) -> int:
        ...

    def __setitem__(self, key: str, value: int) -> None:
        ...
